#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with users.
"""

from __future__ import annotations

from meerschaum.api import debug, CHECK_UPDATE, get_api_connector, no_auth
from meerschaum.api.dash import active_sessions, authenticated_sessions, unauthenticated_sessions
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.utils.typing import WebState, SuccessTuple, List, Tuple, Optional
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
from meerschaum.config import get_config
from meerschaum.core import User
dcc, html = import_dcc(check_update=CHECK_UPDATE), import_html(check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)

def get_users_cards(state: WebState) -> Tuple[List[dbc.Card], List[SuccessTuple]]:
    """
    Return the cards and alerts for users.
    """
    cards, alerts = [], [] 
    conn = get_web_connector(state)
    usernames = conn.get_users(debug=debug)
    for username in usernames:
        cards.append(
            dbc.Card([
                dbc.CardHeader(),
                dbc.CardBody([html.H5(username)]),
            ])
        )

    return cards, alerts


def is_session_authenticated(session_id: str) -> bool:
    """
    Check is a session ID is active.
    If running in secure mode, check whether a session ID corresponds to an admin.

    Parameters
    ----------
    session_id: str
        The session UUID.

    Returns
    -------
    A bool whether the session is authenticated to perform actions.
    """
    if no_auth:
        return True
    if session_id in unauthenticated_sessions:
        return False
    if session_id in authenticated_sessions:
        return True
    permissions = get_config('system', 'api', 'permissions')
    allow_non_admin = permissions.get('actions', {}).get('non_admin', False)
    if allow_non_admin:
        return True
    conn = get_api_connector()
    username = active_sessions.get(session_id, {}).get('username', None)
    user = User(username, instance=conn)
    user_type = conn.get_user_type(user, debug=debug)
    is_auth = user_type == 'admin'
    if is_auth:
        authenticated_sessions[session_id] = username
    else:
        unauthenticated_sessions[session_id] = username
    return is_auth



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This package includes argument parsing utilities.
"""

from meerschaum._internal.arguments._parse_arguments import (
    parse_arguments, parse_line, remove_leading_action,
    parse_dict_to_sysargs,
)
from meerschaum._internal.arguments._parser import parser
from meerschaum.plugins import add_plugin_argument

__all__ = [
    'parser', 'parse_arguments', 'parse_line', 'add_plugin_argument', 'parse_dict_to_sysargs',
    'remove_leading_action',
]



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Create an APIConnector from a URI string.
"""

from meerschaum.utils.typing import Optional, Dict, Any, Union
from meerschaum.utils.warnings import warn, error

@classmethod
def from_uri(
        cls,
        uri: str,
        label: Optional[str] = None,
        as_dict: bool = False,
    ) -> Union[
        'meerschaum.connectors.APIConnector',
        Dict[str, Union[str, int]],
    ]:
    """
    Create a new APIConnector from a URI string.

    Parameters
    ----------
    uri: str
        The URI connection string.

    label: Optional[str], default None
        If provided, use this as the connector label.
        Otherwise use the determined database name.

    as_dict: bool, default False
        If `True`, return a dictionary of the keyword arguments
        necessary to create a new `APIConnector`, otherwise create a new object.

    Returns
    -------
    A new APIConnector object or a dictionary of attributes (if `as_dict` is `True`).
    """
    from meerschaum.connectors.sql import SQLConnector
    params = SQLConnector.parse_uri(uri)
    if 'host' not in params:
        error("No host was found in the provided URI.")
    params['protocol'] = params.pop('flavor')
    params['label'] = label or (
        (
            (params['username'] + '@' if 'username' in params else '')
            + params['host']
        ).lower()
    )

    return cls(**params) if not as_dict else params



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with pipes via the web interface.
"""

from __future__ import annotations
import json
import shlex
from textwrap import dedent
from dash.dependencies import Input, Output, State
from meerschaum.utils.typing import List, Optional, Dict, Any, Tuple, Union
from meerschaum.utils.misc import string_to_dict, json_serialize_datetime
from meerschaum.utils.packages import attempt_import, import_dcc, import_html, import_pandas
from meerschaum.utils.sql import get_pd_type
from meerschaum.utils.yaml import yaml
from meerschaum.connectors.sql._fetch import get_pipe_query
from meerschaum.api import endpoints, CHECK_UPDATE
from meerschaum.api.dash import (
    dash_app, debug, _get_pipes
)
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.api.dash.components import alert_from_success_tuple
from meerschaum.api.dash.users import is_session_authenticated
import meerschaum as mrsm
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
dash_ace = attempt_import('dash_ace', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
humanfriendly = attempt_import('humanfriendly', check_update=CHECK_UPDATE)
pd = import_pandas()

def pipe_from_ctx(ctx, trigger_property: str = 'n_clicks') -> Union[mrsm.Pipe, None]:
    """
    Return a `meerschaum.Pipe` object from a dynamic object with an
    index of a pipe's meta dictionary.
    """
    try:
        ### I know this looks confusing and feels like a hack.
        ### Because Dash JSON-ifies the ID dictionary and we are including a JSON-ified dictionary,
        ### we have to do some crazy parsing to get the pipe's meta-dict back out of it
        meta = json.loads(json.loads(ctx[0]['prop_id'].split('.' + trigger_property)[0])['index'])
    except Exception as e:
        meta = None
    if meta is None:
        return None
    return mrsm.Pipe(**meta)

def keys_from_state(
        state: Dict[str, Any],
        with_params: bool = False
    ) -> Union[
        Tuple[List[str], List[str], List[str]],
        Tuple[List[str], List[str], List[str], str],
    ]:
    """
    Read the current state and return the selected keys lists.
    """
    _filters = {
        'ck' : state.get(f"connector-keys-{state['pipes-filter-tabs.active_tab']}.value", None),
        'mk' : state.get(f"metric-keys-{state['pipes-filter-tabs.active_tab']}.value", None),
        'lk' : state.get(f"location-keys-{state['pipes-filter-tabs.active_tab']}.value", None),
    }
    if state['pipes-filter-tabs.active_tab'] == 'input':
        try:
            #  params = string_to_dict(state['params-textarea.value'])
            params = string_to_dict(state['search-parameters-editor.value'])
        except Exception as e:
            params = None
    else:
        params = None

    for k in _filters:
        _filters[k] = [] if _filters[k] is None else _filters[k]
        if not isinstance(_filters[k], list):
            try:
                _filters[k] = shlex.split(_filters[k])
            except Exception as e:
                print(e)
                _filters[k] = []
    keys = [_filters['ck'], _filters['mk'], _filters['lk']]
    if with_params:
        keys.append(params)
    return tuple(keys)

def pipes_from_state(
        state: Dict[str, Any],
        **kw
    ):
    _ck, _mk, _lk, _params = keys_from_state(state, with_params=True)
    try:
        _pipes = _get_pipes(
            _ck, _mk, _lk,
            params = _params,
            mrsm_instance = get_web_connector(state), 
            **kw
        )
    except Exception as e:
        return False, str(e)
    return _pipes

def get_pipes_cards(*keys, session_data: Optional[Dict[str, Any]] = None):
    """
    Returns a tuple:
        - pipes as a list of cards
        - alert list
    """
    cards = []
    session_id = (session_data or {}).get('session-id', None)
    authenticated = is_session_authenticated(str(session_id))

    _pipes = pipes_from_state(*keys, as_list=True)
    alerts = [alert_from_success_tuple(_pipes)]
    if not isinstance(_pipes, list):
        _pipes = []
    for p in _pipes:
        meta_str = json.dumps(p.meta)
        footer_children = dbc.Row(
            [
                dbc.Col(
                    (
                        dbc.DropdownMenu(
                            label = "Manage",
                            children = [
                                dbc.DropdownMenuItem(
                                    'Sync',
                                    id = {
                                        'type': 'manage-pipe-button',
                                        'index': meta_str,
                                        'action': 'sync',
                                    },
                                ),
                                dbc.DropdownMenuItem(
                                    'Drop',
                                    id = {
                                        'type': 'manage-pipe-button',
                                        'index': meta_str,
                                        'action': 'drop',
                                    },
                                ),
                                dbc.DropdownMenuItem(
                                    'Delete',
                                    id = {
                                        'type': 'manage-pipe-button',
                                        'index': meta_str,
                                        'action': 'delete',
                                    },
                                ),
                            ],
                            direction = "end",
                            menu_variant = "dark",
                            size = 'sm',
                            color = 'secondary',
                        )
                    ) if authenticated else [],
                    width = 2,
                ),
                #  dbc.Col(
                    #  (
                        #  dbc.Button(
                            #  'Sync',
                            #  size = 'sm',
                            #  style = {'width': '100%'},
                            #  id = {'type': 'pipe-sync-button', 'index': meta_str},
                        #  ) if authenticated else []
                    #  ),
                    #  width = 2,
                #  ),
                dbc.Col(width=6),
                dbc.Col(
                    dbc.Button(
                        'Download recent data',
                        size = 'sm',
                        color = 'link',
                        style = {'float': 'right'},
                        id = {'type': 'pipe-download-csv-button', 'index': meta_str},
                    ),
                    width = 4,
                ),
            ],
            justify = 'start',
        )
        card_body_children = [
            html.H5(
                html.B(str(p)),
                className = 'card-title',
                style = {'font-family': ['monospace']}
            ),
            html.Div(
                dbc.Accordion(
                    accordion_items_from_pipe(p, authenticated=authenticated),
                    flush = True,
                    start_collapsed = True,
                    id = {'type': 'pipe-accordion', 'index': meta_str},
                )
            )

        ]
        cards.append(
            dbc.Card(children=[
                dbc.CardBody(children=card_body_children),
                dbc.CardFooter(children=footer_children),
            ])
        )
    return cards, alerts


def accordion_items_from_pipe(
        pipe: mrsm.Pipe,
        active_items: Optional[List[str]] = None,
        authenticated: bool = False,
    ) -> List[dbc.AccordionItem]:
    """
    Build the accordion items for a given pipe.
    """
    if active_items is None:
        active_items = []

    items_titles = {
        'overview': '🔑 Keys',
        'stats': '🧮 Statistics',
        'columns': '🏛️ Columns',
        'parameters': '📔 Parameters',
    }
    if pipe.connector_keys.startswith('sql:'):
        items_titles['sql'] = '📃 SQL Query'
    items_titles.update({
        'recent-data': '🗃️ Recent Data',
        'sync-data': '📝 Sync Documents',
    })

    skip_items = (
        ['sync-data']
        if not authenticated
        else []
    )
    for item in skip_items:
        _ = items_titles.pop(item, None)

    ### Only generate items if they're in the `active_items` list.
    items_bodies = {}
    if 'overview' in active_items:
        overview_header = [html.Thead(html.Tr([html.Th("Attribute"), html.Th("Value")]))]
        dt_name, id_name, val_name = pipe.get_columns('datetime', 'id', 'value', error=False)
        overview_rows = [
            html.Tr([html.Td("Connector"), html.Td(f"{pipe.connector_keys}")]),
            html.Tr([html.Td("Metric"), html.Td(f"{pipe.metric_key}")]),
            html.Tr([html.Td("Location"), html.Td(f"{pipe.location_key}")]),
            html.Tr([html.Td("Instance"), html.Td(f"{pipe.instance_keys}")]),
            html.Tr([html.Td("Target Table"), html.Td(f"{pipe.target}")]),
        ]
        for col_key, col in pipe.columns.items():
            overview_rows.append(html.Tr([html.Td(f"'{col_key}' Index"), html.Td(col)]))
        tags = pipe.tags
        if tags:
            tags_items = html.Ul([
                html.Li(tag)
                for tag in tags
            ])
            overview_rows.append(
                html.Tr([
                    html.Td("Tags"),
                    html.Td(tags_items),
                ])
            )

        items_bodies['overview'] = dbc.Table(
            overview_header + [html.Tbody(overview_rows)],
            bordered=False, hover=True, striped=False,
        )

    if 'stats' in active_items:
        stats_header = [html.Thead(html.Tr([html.Th("Statistic"), html.Th("Value")]))]
        try:
            oldest_time = pipe.get_sync_time(newest=False, round_down=False, debug=debug)
            newest_time = pipe.get_sync_time(newest=True, round_down=False, debug=debug)
            interval = (
                (newest_time - oldest_time) if newest_time is not None and oldest_time is not None
                else None
            )
            rowcount = pipe.get_rowcount(debug=debug)
        except Exception as e:
            oldest_time = None
            newest_time = None
            interval = None
            rowcount = None

        stats_rows = []
        if rowcount is not None:
            stats_rows.append(html.Tr([html.Td("Row-count"), html.Td(f"{rowcount}")]))
        if interval is not None:
            stats_rows.append(
                html.Tr([html.Td("Timespan"), html.Td(humanfriendly.format_timespan(interval))])
            )
        if oldest_time is not None:
            stats_rows.append(html.Tr([html.Td("Oldest time"), html.Td(str(oldest_time))]))
        if newest_time is not None:
            stats_rows.append(html.Tr([html.Td("Newest time"), html.Td(str(newest_time))]))


        items_bodies['stats'] = dbc.Table(stats_header + [html.Tbody(stats_rows)], hover=True)

    if 'columns' in active_items:
        try:
            columns_header = [html.Thead(html.Tr([
                html.Th("Column"), html.Th("DB Type"), html.Th("PD Type")
            ]))]
            columns_types = {
                col: typ.replace('_', ' ')
                for col, typ in pipe.get_columns_types(debug=debug).items()
            }
            columns_rows = [
                html.Tr([
                    html.Td(html.Pre(col)),
                    html.Td(html.Pre(typ)),
                    html.Td(html.Pre(get_pd_type(typ))),
                ]) for col, typ in columns_types.items()
            ]
            columns_body = [html.Tbody(columns_rows)]
            columns_table = dbc.Table(columns_header + columns_body, bordered=False, hover=True)
        except Exception as e:
            columns_table = html.P("Could not retrieve columns ― please try again.")
        items_bodies['columns'] = columns_table

    if 'parameters' in active_items:
        parameters_editor = dash_ace.DashAceEditor(
            value = yaml.dump(pipe.parameters),
            mode = 'norm',
            tabSize = 4,
            theme = 'twilight',
            id = {'type': 'parameters-editor', 'index': json.dumps(pipe.meta)},
            width = '100%',
            height = '500px',
            readOnly = False,
            showGutter = True,
            showPrintMargin = True,
            highlightActiveLine = True,
            wrapEnabled = True,
            style = {'min-height': '120px'},
        )
        update_parameters_button = dbc.Button(
            "Update",
            id = {'type': 'update-parameters-button', 'index': json.dumps(pipe.meta)},
        )

        as_yaml_button = dbc.Button(
            "YAML",
            id = {'type': 'parameters-as-yaml-button', 'index': json.dumps(pipe.meta)},
            color = 'link',
            size = 'sm',
            style = {'text-decoration': 'none'},
        )
        as_json_button = dbc.Button(
            "JSON",
            id = {'type': 'parameters-as-json-button', 'index': json.dumps(pipe.meta)},
            color = 'link',
            size = 'sm',
            style = {'text-decoration': 'none', 'margin-left': '10px'},
        )
        items_bodies['parameters'] = html.Div([
            parameters_editor,
            html.Br(),
            dbc.Row([
                dbc.Col(html.Span(
                    (
                        ([update_parameters_button] if authenticated else []) +
                        [
                            as_json_button,
                            as_yaml_button,
                        ]
                    )
                ), width=4),
                dbc.Col([
                    html.Div(
                        id={'type': 'update-parameters-success-div', 'index': json.dumps(pipe.meta)}
                    )
                ],
                width=True,
                )
            ]),

        ])

    if 'sql' in active_items:
        query = dedent((get_pipe_query(pipe, warn=False) or "")).lstrip().rstrip()
        sql_editor = dash_ace.DashAceEditor(
            value = query,
            mode = 'sql',
            tabSize = 4,
            theme = 'twilight',
            id = {'type': 'sql-editor', 'index': json.dumps(pipe.meta)},
            width = '100%',
            height = '500px',
            readOnly = False,
            showGutter = True,
            showPrintMargin = False,
            highlightActiveLine = True,
            wrapEnabled = True,
            style = {'min-height': '120px'},
        )
        update_sql_button = dbc.Button(
            "Update",
            id = {'type': 'update-sql-button', 'index': json.dumps(pipe.meta)},
        )
        items_bodies['sql'] = html.Div([
            sql_editor,
            html.Br(),
            dbc.Row([
                dbc.Col([update_sql_button], width=2),
                dbc.Col([
                    html.Div(
                        id={'type': 'update-sql-success-div', 'index': json.dumps(pipe.meta)}
                    )
                ],
                width=True,
                )
            ]),
        ])

    if 'recent-data' in active_items:
        try:
            df = pipe.get_backtrack_data(backtrack_minutes=10, limit=10, debug=debug).astype(str)
            table = dbc.Table.from_dataframe(df, bordered=False, hover=True) 
        except Exception as e:
            table = html.P("Could not retrieve recent data.")
        items_bodies['recent-data'] = table

    if 'sync-data' in active_items:
        backtrack_df = pipe.get_backtrack_data(debug=debug, limit=1)
        try:
            json_text = backtrack_df.fillna(pd.NA).to_json(
                orient = 'records',
                date_format = 'iso',
                force_ascii = False,
                indent = 4,
                date_unit = 'ns',
            ) if backtrack_df is not None else '[]'
        except Exception as e:
            warn(e)
            json_text = '[]'

        json_text = json.dumps(json.loads(json_text), indent=4, separators=(',', ': '))
        sync_editor = dash_ace.DashAceEditor(
            value = json_text,
            mode = 'norm',
            tabSize = 4,
            theme = 'twilight',
            id = {'type': 'sync-editor', 'index': json.dumps(pipe.meta)},
            width = '100%',
            height = '500px',
            readOnly = False,
            showGutter = True,
            showPrintMargin = False,
            highlightActiveLine = True,
            wrapEnabled = True,
            style = {'min-height': '120px'},
        )
        update_sync_button = dbc.Button(
            "Sync",
            id = {'type': 'update-sync-button', 'index': json.dumps(pipe.meta)},
        )
        sync_success_div = html.Div(id={'type': 'sync-success-div', 'index': json.dumps(pipe.meta)})
        items_bodies['sync-data'] = html.Div([
            sync_editor,
            html.Br(),
            dbc.Row([
                dbc.Col([update_sync_button], width=1),
                dbc.Col([sync_success_div], width=True),
            ]),
        ])

    return [
        dbc.AccordionItem(items_bodies.get(item_id, ''), title=title, item_id=item_id)
        for item_id, title in items_titles.items()
    ]




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Upgrade your current Meerschaum environment
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any, List, Optional, Union

def upgrade(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Upgrade Meerschaum, plugins, or packages.
    
    Command:
        `upgrade {option}`
    
    Example:
        `upgrade meerschaum`
    """

    from meerschaum.actions import choose_subaction
    options = {
        'plugins'    : _upgrade_plugins,
        'meerschaum' : _upgrade_meerschaum,
        'mrsm'       : _upgrade_meerschaum,
        'packages'   : _upgrade_packages,
    }
    return choose_subaction(action, options, **kw)


def _upgrade_meerschaum(
        action: Optional[List[str]] = None,
        yes: bool = False,
        force: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Upgrade the current Meerschaum instance.
    Optionally specify dependency groups.
    
    Examples:
        - `upgrade meerschaum`
        - `upgrade meerschaum full`

    """
    import subprocess
    import json
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import info
    from meerschaum.actions import actions
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.packages import pip_install, attempt_import
    from meerschaum.utils.misc import is_docker_available

    if action is None:
        action = []

    is_stack_running = False
    try:
        success, msg = actions['stack'](
            ['ps', '-q'], _capture_output=True, debug=debug
        )
        if msg.endswith('\n'):
            msg = msg[:-1]
        containers = msg.split('\n')
    except Exception as e:
        containers = []

    if containers:
        if force:
            answer = True
        else:
            answer = yes_no(f"Take down the stack?", default='y', yes=yes, noask=noask)

        if answer:
            if debug:
                info("Taking stack down...")
            actions['stack'](['down'], debug=debug)

    dependencies = None
    if action != [''] and len(action) > 0:
        dependencies = action[0]

    install_name = 'meerschaum' + (
        ('[' + ','.join(dependencies.split(',')) + ']') if dependencies else ''
    )

    if debug:
        dprint('Upgrade meerschaum with dependencies: \"' + f'{dependencies}' + '\"')
    if not pip_install(install_name, venv=None, debug=debug):
        return False, f"Failed to upgrade Meerschaum via pip."

    if debug:
        dprint("Pulling new Docker images...")
    if is_docker_available():
        actions['stack'](['pull'], debug=debug)

    return True, "Success"


class NoVenv:
    pass

def _upgrade_packages(
        action: Optional[List[str]] = None,
        venv: Union[str, None, NoVenv] = NoVenv,
        yes: bool = False,
        force: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Upgrade and install dependencies.
    If provided, upgrade only a dependency group, otherwise default to `full`.
    
    Examples:
        ```
        upgrade packages
        ```
    
        ```
        upgrade packages docs
        ```
    """
    from meerschaum.utils.packages import packages, pip_install
    from meerschaum.utils.warnings import info, warn
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.formatting import make_header, pprint
    from meerschaum.utils.misc import print_options
    if action is None:
        action = []
    if venv is NoVenv:
        venv = 'mrsm'
    if len(action) == 0:
        group = 'full'
    else:
        group = action[0]

    if group not in packages:
        invalid_msg = f"Invalid dependency group '{group}'."
        avail_msg = make_header("Available groups:")
        for k in packages:
            avail_msg += "\n  - {k}"

        warn(invalid_msg + "\n\n" + avail_msg, stack=False)
        return False, invalid_msg

    print(make_header("Packages to Upgrade:"))
    pprint(packages[group])
    question = (
        f"Are you sure you want to upgrade {len(packages[group])} packages " +
        f"(dependency group '{group}')?"
    )
    to_install = [install_name for import_name, install_name in packages[group].items()]

    success, msg = False, f"Nothing installed."
    if force or yes_no(question, noask=noask, yes=yes):
        success = pip_install(*to_install, debug=debug)
        msg = (
            f"Successfully installed {len(packages[group])} packages." if success
            else f"Failed to install packages in dependency group '{group}'."
        )
    return success, msg

def _upgrade_plugins(
        action: Optional[List[str]] = None,
        yes: bool = False,
        force: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Upgrade all installed plugins to the latest versions.
    If no plugins are specified, attempt to upgrade all,
    otherwise only upgrade the specified plugins.
    
    Examples:
    
    ```
    upgrade plugins
    ```
    ```
    upgrade plugins testing
    ```
    """
    from meerschaum.actions import actions
    from meerschaum.plugins import get_plugins_names
    from meerschaum.utils.misc import print_options
    from meerschaum.utils.prompt import yes_no

    if action is None:
        action = []

    to_install = get_plugins_names() if len(action) == 0 else action
    if len(to_install) == 0:
        return False, "No plugins to upgrade."
    print_options(to_install, header="Plugins to Upgrade:")
    if force or yes_no(f"Upgrade {len(to_install)} plugins?", yes=yes, noask=noask):
        return actions['install'](
            action=['plugins'] + to_install, debug=debug, force=force, noask=noask, yes=yes, **kw
        )
    return False, "No plugins upgraded."

### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
upgrade.__doc__ += _choices_docstring('upgrade')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
NOTE: This action may be a huge security vulnerability
    if not handled appropriately!
"""

from meerschaum.utils.typing import SuccessTuple, List, Any, Optional

def sh(
        action: Optional[List[str]] = None,
        sub_args: Optional[List[str]] = None,
        filtered_sysargs: Optional[List[str]] = None,
        use_bash: bool = True,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Execute system commands.
    """
    import subprocess
    import sys, os, shlex
    from meerschaum.utils.warnings import error
    from meerschaum.utils.debug import dprint

    if action is None:
        action = []
    if sub_args is None:
        sub_args = []
    if filtered_sysargs is None:
        filtered_sysargs = []

    _shell = os.environ.get('SHELL', 'bash')

    if action:
        if action[0].startswith('!'):
            action[0] = action[0][1:]
        elif action[0] == 'sh':
            action = action[1:]
    cmd_list = action

    if sub_args:
        cmd_list += sub_args

    command_list = cmd_list
    if use_bash:
        command_list = ["bash"]
        if len(action) != 0:
            try:
                command_list += ["-c", shlex.join(cmd_list)]
            except Exception as e:
                command_list += ["-c", ' '.join(cmd_list)]
    else:
        if len(action) == 0:
            command_list = _shell

    if debug:
        dprint(f'action  : {action}')
        dprint(f'sub-args: {sub_args}')
        dprint(command_list)

    try:
        process = subprocess.Popen(
            command_list,
            shell = False,
            env = os.environ,
        )
        exit_code = process.wait()
    except FileNotFoundError:
        msg = f"Invalid commands: '{command_list}'"
        return False, msg
    except KeyboardInterrupt:
        return True, "Success"

    if exit_code != 0:
        return (False, f"Returned exit code: {exit_code}")
    return True, "Success"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Fetch and manipulate Pipes' attributes
"""

from __future__ import annotations
from meerschaum.utils.typing import Tuple, Dict, SuccessTuple, Any, Union, Optional, List
from meerschaum.utils.warnings import warn

@property
def attributes(self) -> Dict[str, Any]:
    """
    Return a dictionary of a pipe's keys and parameters.
    These values are reflected directly from the pipes table of the instance.
    """
    import time
    from meerschaum.config import get_config
    from meerschaum.config._patch import apply_patch_to_config
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    timeout_seconds = get_config('pipes', 'attributes', 'local_cache_timeout_seconds')

    if '_attributes' not in self.__dict__:
        self._attributes = {}

    now = time.perf_counter()
    last_refresh = self.__dict__.get('_attributes_sync_time', None)
    timed_out = (
        last_refresh is None
        or
        (timeout_seconds is not None and (now - last_refresh) >= timeout_seconds)
    )
    if not self.temporary and timed_out:
        self._attributes_sync_time = now
        local_attributes = self.__dict__.get('_attributes', {})
        with Venv(get_connector_plugin(self.instance_connector)):
            instance_attributes = self.instance_connector.get_pipe_attributes(self)
        self._attributes = apply_patch_to_config(instance_attributes, local_attributes)
    return self._attributes


@property
def parameters(self) -> Optional[Dict[str, Any]]:
    """
    Return the parameters dictionary of the pipe.
    """
    if 'parameters' not in self.attributes:
        self.attributes['parameters'] = {}
    return self.attributes['parameters']


@parameters.setter
def parameters(self, parameters: Dict[str, Any]) -> None:
    """
    Set the parameters dictionary of the in-memory pipe.
    Call `meerschaum.Pipe.edit()` to persist changes.
    """
    self.attributes['parameters'] = parameters


@property
def columns(self) -> Union[Dict[str, str], None]:
    """
    Return the `columns` dictionary defined in `meerschaum.Pipe.parameters`.
    """
    if 'columns' not in self.parameters:
        self.parameters['columns'] = {}
    cols = self.parameters['columns']
    if not isinstance(cols, dict):
        cols = {}
        self.parameters['columns'] = cols
    return cols


@columns.setter
def columns(self, columns: Dict[str, str]) -> None:
    """
    Override the columns dictionary of the in-memory pipe.
    Call `meerschaum.Pipe.edit()` to persist changes.
    """
    if not isinstance(columns, dict):
        warn(f"{self}.columns must be a dictionary, received {type(columns)}")
        return
    self.parameters['columns'] = columns


@property
def tags(self) -> Union[List[str], None]:
    """
    If defined, return the `tags` list defined in `meerschaum.Pipe.parameters`.
    """
    if 'tags' not in self.parameters:
        self.parameters['tags'] = []
    return self.parameters['tags']


@tags.setter
def tags(self, _tags: List[str, str]) -> None:
    """
    Override the tags list of the in-memory pipe.
    Call `meerschaum.Pipe.edit` to persist changes.
    """
    from meerschaum.utils.warnings import error
    from meerschaum.config.static import STATIC_CONFIG
    negation_prefix = STATIC_CONFIG['system']['fetch_pipes_keys']['negation_prefix']
    for t in _tags:
        if t.startswith(negation_prefix):
            error(f"Tags cannot begin with '{negation_prefix}'.")
    self.parameters['tags'] = _tags


@property
def dtypes(self) -> Union[Dict[str, Any], None]:
    """
    If defined, return the `dtypes` dictionary defined in `meerschaum.Pipe.parameters`.
    """
    from meerschaum.config._patch import apply_patch_to_config
    configured_dtypes = self.parameters.get('dtypes', {})
    remote_dtypes = self.infer_dtypes(persist=False)
    patched_dtypes = apply_patch_to_config(remote_dtypes, configured_dtypes)
    self.parameters['dtypes'] = patched_dtypes
    return self.parameters['dtypes']


@dtypes.setter
def dtypes(self, _dtypes: Dict[str, Any]) -> None:
    """
    Override the dtypes dictionary of the in-memory pipe.
    Call `meerschaum.Pipe.edit()` to persist changes.
    """
    self.parameters['dtypes'] = _dtypes


def get_columns(self, *args: str, error: bool = False) -> Union[str, Tuple[str]]:
    """
    Check if the requested columns are defined.

    Parameters
    ----------
    *args: str
        The column names to be retrieved.
        
    error: bool, default False
        If `True`, raise an `Exception` if the specified column is not defined.

    Returns
    -------
    A tuple of the same size of `args` or a `str` if `args` is a single argument.

    Examples
    --------
    >>> pipe = mrsm.Pipe('test', 'test')
    >>> pipe.columns = {'datetime': 'dt', 'id': 'id'}
    >>> pipe.get_columns('datetime', 'id')
    ('dt', 'id')
    >>> pipe.get_columns('value', error=True)
    Exception:  🛑 Missing 'value' column for Pipe('test', 'test').
    """
    from meerschaum.utils.warnings import error as _error, warn
    if not args:
        args = tuple(self.columns.keys())
    col_names = []
    for col in args:
        col_name = None
        try:
            col_name = self.columns[col]
            if col_name is None and error:
                _error(f"Please define the name of the '{col}' column for {self}.")
        except Exception as e:
            col_name = None
        if col_name is None and error:
            _error(f"Missing '{col}'" + f" column for {self}.")
        col_names.append(col_name)
    if len(col_names) == 1:
        return col_names[0]
    return tuple(col_names)


def get_columns_types(self, debug: bool = False) -> Union[Dict[str, str], None]:
    """
    Get a dictionary of a pipe's column names and their types.

    Parameters
    ----------
    debug: bool, default False:
        Verbosity toggle.

    Returns
    -------
    A dictionary of column names (`str`) to column types (`str`).

    Examples
    --------
    >>> pipe.get_columns_types()
    {
      'dt': 'TIMESTAMP WITHOUT TIMEZONE',
      'id': 'BIGINT',
      'val': 'DOUBLE PRECISION',
    }
    >>>
    """
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    with Venv(get_connector_plugin(self.instance_connector)):
        return self.instance_connector.get_pipe_columns_types(self, debug=debug)


def get_id(self, **kw: Any) -> Union[int, None]:
    """
    Fetch a pipe's ID from its instance connector.
    If the pipe does not exist, return `None`.
    """
    if self.temporary:
        return None
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    with Venv(get_connector_plugin(self.instance_connector)):
        return self.instance_connector.get_pipe_id(self, **kw)


@property
def id(self) -> Union[int, None]:
    """
    Fetch and cache a pipe's ID.
    """
    if not ('_id' in self.__dict__ and self._id):
        self._id = self.get_id()
    return self._id


def get_val_column(self, debug: bool = False) -> Union[str, None]:
    """
    Return the name of the value column if it's defined, otherwise make an educated guess.
    If not set in the `columns` dictionary, return the first numeric column that is not
    an ID or datetime column.
    If none may be found, return `None`.

    Parameters
    ----------
    debug: bool, default False:
        Verbosity toggle.

    Returns
    -------
    Either a string or `None`.
    """
    from meerschaum.utils.debug import dprint
    if debug:
        dprint('Attempting to determine the value column...')
    try:
        val_name = self.get_columns('value')
    except Exception as e:
        val_name = None
    if val_name is not None:
        if debug:
            dprint(f"Value column: {val_name}")
        return val_name

    cols = self.columns
    if cols is None:
        if debug:
            dprint('No columns could be determined. Returning...')
        return None
    try:
        dt_name = self.get_columns('datetime', error=False)
    except Exception as e:
        dt_name = None
    try:
        id_name = self.get_columns('id', errors=False)
    except Exception as e:
        id_name = None

    if debug:
        dprint(f"dt_name: {dt_name}")
        dprint(f"id_name: {id_name}")

    cols_types = self.get_columns_types(debug=debug)
    if cols_types is None:
        return None
    if debug:
        dprint(f"cols_types: {cols_types}")
    if dt_name is not None:
        cols_types.pop(dt_name, None)
    if id_name is not None:
        cols_types.pop(id_name, None)

    candidates = []
    candidate_keywords = {'float', 'double', 'precision', 'int', 'numeric',}
    for search_term in candidate_keywords:
        for col, typ in cols_types.items():
            if search_term in typ.lower():
                candidates.append(col)
                break
    if not candidates:
        if debug:
            dprint(f"No value column could be determined.")
        return None

    return candidates[0]


@property
def parents(self) -> List[meerschaum.Pipe]:
    """
    Return a list of `meerschaum.Pipe` objects to be designated as parents.
    """
    if 'parents' not in self.parameters:
        return []
    from meerschaum.utils.warnings import warn
    _parents_keys = self.parameters['parents']
    if not isinstance(_parents_keys, list):
        warn(
            f"Please ensure the parents for {self} are defined as a list of keys.",
            stacklevel = 4
        )
        return []
    from meerschaum import Pipe
    _parents = []
    for keys in _parents_keys:
        try:
            p = Pipe(**keys)
        except Exception as e:
            warn(f"Unable to build parent with keys '{keys}' for {self}:\n{e}")
            continue
        _parents.append(p)
    return _parents


@property
def children(self) -> List[meerschaum.Pipe]:
    """
    Return a list of `meerschaum.Pipe` objects to be designated as children.
    """
    if 'children' not in self.parameters:
        return []
    from meerschaum.utils.warnings import warn
    _children_keys = self.parameters['children']
    if not isinstance(_children_keys, list):
        warn(
            f"Please ensure the children for {self} are defined as a list of keys.",
            stacklevel = 4
        )
        return []
    from meerschaum import Pipe
    _children = []
    for keys in _children_keys:
        try:
            p = Pipe(**keys)
        except Exception as e:
            warn(f"Unable to build parent with keys '{keys}' for {self}:\n{e}")
            continue
        _children.append(p)
    return _children


@property
def target(self) -> str:
    """
    The target table name.
    You can set the target name under on of the following keys
    (checked in this order):
      - `target`
      - `target_name`
      - `target_table`
      - `target_table_name`
    """
    if 'target' not in self.parameters:
        target = self._target_legacy()
        potential_keys = ('target_name', 'target_table', 'target_table_name')
        for k in potential_keys:
            if k in self.parameters:
                target = self.parameters[k]
                break

        if self.instance_connector.type == 'sql':
            from meerschaum.utils.sql import truncate_item_name
            truncated_target = truncate_item_name(target, self.instance_connector.flavor)
            if truncated_target != target:
                warn(
                    f"The target '{target}' is too long for '{self.instance_connector.flavor}', "
                    + f"will use {truncated_target} instead."
                )
                target = truncated_target

        self.target = target
    return self.parameters['target']


def _target_legacy(self) -> str:
    """
    The old method of determining a pipe's table name by joining the keys with underscores.
    **NOTE:** Converts the `':'` in the `connector_keys` to an `'_'`.
    """
    name = f"{self.connector_keys.replace(':', '_')}_{self.metric_key}"
    if self.location_key is not None:
        name += f"_{self.location_key}"
    return name


@target.setter
def target(self, _target: str) -> None:
    """
    Override the target of the in-memory pipe.
    Call `meerschaum.Pipe.edit` to persist changes.
    """
    self.parameters['target'] = _target


def guess_datetime(self) -> Union[str, None]:
    """
    Try to determine a pipe's datetime column.
    """
    dtypes = self.dtypes

    ### Abort if the user explictly disallows a datetime index.
    if 'datetime' in dtypes:
        if dtypes['datetime'] is None:
            return None

    dt_cols = [
        col for col, typ in self.dtypes.items()
        if str(typ).startswith('datetime')
    ]
    if not dt_cols:
        return None
    return dt_cols[0]    


def get_indices(self) -> Dict[str, str]:
    """
    Return a dictionary in the form of `pipe.columns` but map to index names.
    """
    return {
        ix: (self.target + '_' + col + '_index')
        for ix, col in self.columns.items() if col
    }



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Directly interact with the SQL server via Meerschaum actions
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Any, List, Optional

exec_methods = {
    'read',
    'exec',
}

def sql(
        action: Optional[List[str]] = None,
        gui: bool = False,
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ):
    """Execute a SQL query or launch an interactive CLI. All positional arguments are optional.
    
    Usage:
        `sql {label} {method} {query / table}`
    
    Options:
        - `sql {label}`
            Launch an interactive CLI. If {label} is omitted, use 'main'.
    
        - `sql {label} read [query / table]`
            Read a table or query as a pandas DataFrame and print the result.
    
        - `sql {label} exec [query]`
            Execute a query and print the success status.
    
    Examples:
        - `sql`
            Open an interactive CLI for `sql:main`.
    
        - `sql local`
            Open an interactive CLI for `sql:local`.
    
        - `sql table`
            Read from `table` on `sql:main`
              (translates to `SELECT * FROM table`).
    
        - `sql local table`
            Read from `table` on `sql:local`.
    
        - `sql local read table`
            Read from `table` on `sql:local`.
    
        - `sql "SELECT * FROM table WHERE id = 1"`
            Execute the above query on `sql:main` and print the results.
    
        - `sql local exec "INSERT INTO table (id) VALUES (1)"
            Execute the above query on `sql:local`.
    """

    if action is None:
        action = []

    ### input: `sql read` or `sql exec`
    if len(action) == 1 and action[0] in exec_methods:
        import textwrap
        print(textwrap.dedent(sql.__doc__))
        return (False, 'Query must not be empty.')

    from meerschaum.connectors import get_connector

    ### input: `sql`
    if len(action) == 0:
        return get_connector('sql').cli(debug=debug) 

    from meerschaum.config import get_config
    from meerschaum.utils.debug import dprint

    method = None
    label = None
    query = action[-1]
    for a in action[:]:
        ### check if user specifies an exec method (read vs exec)
        if a in exec_methods:
            method = a
            continue
        ### check if user specifies a label
        if a in get_config('meerschaum', 'connectors', 'sql', patch=True):
            label = a
    if method is None:
        method = 'read'
    if label is None:
        label = 'main'

    conn = get_connector(type='sql', label=label, debug=debug)
    if not conn:
        return False, (f"Could not create SQL connector '{label}'.\n"
            f"Verify meerschaum:connectors:sql:{label} is defined correctly with `show connectors`," + 
            " and update or add connectors with `edit config`."
        )

    if debug:
        dprint(action)
        dprint(conn)

    ### input: `sql main`
    if query == label:
        if debug:
            dprint(f"No query provided. Opening CLI on connector '{label}'")
        return conn.cli(debug=debug)

    ### guess the method from the structure of the query
    if 'select' in query.lower() or ' ' not in query:
        if method != 'read':
            method = 'read'
    else:
        if method != 'exec':
            method = 'exec'

    result = getattr(conn, method)(query, debug=debug)
    if result is False:
        return (False, f"Failed to execute query:\n\n{query}")
    
    from meerschaum.utils.packages import attempt_import, import_pandas
    from meerschaum.utils.formatting import print_tuple, pprint
    sqlalchemy_engine_result = attempt_import('sqlalchemy.engine.result')
    pd = import_pandas()
    if 'sqlalchemy' in str(type(result)):
        if not nopretty:
            print_tuple((True, f"Successfully executed query:\n\n{query}"))
    else:
        if not nopretty:
            pprint(result)
        else:
            print(
                result.fillna(pd.NA).to_json(
                    date_format = 'iso',
                    orient = 'split',
                    index = False,
                    date_unit = 'us'
                )
            )

    return (True, "Success")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define components for choosing keys.
"""

from __future__ import annotations
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
from meerschaum.actions import actions
from meerschaum.api import CHECK_UPDATE
from meerschaum.api.dash.components import search_parameters_editor
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)

placeholders = {
    'ck' : 'Connectors',
    'mk' : 'Metrics',
    'lk' : 'Locations',
    'params' : (
        'Additional search parameters. ' +
        'Simple dictionary format or JSON accepted.'
    ),
}
widths = {
    'flags' : {'size' : 12},
    'params' : {'size' : 12},
    'begin_end' : {'size' : 8},
    'arguments' : {'size' : 8},
}
input_group_sizes = {
    'ck' : 'sm',
    'mk' : 'sm',
    'lk' : 'sm',
    'params' : 'sm',
}

action_dropdown_row = html.Div(
        children = [
        dbc.Row(
            id = 'action-row',
            children = [
                dbc.Col(
                    html.Div(
                        dbc.Select(
                            id = 'action-dropdown',
                            options = [],
                            className = 'custom-select input-text'
                            #  clearable = False,
                            #  style = {'height' : '150px'},
                        ),
                        id = 'action-dropdown-div',
                        className = 'dbc_dark',
                    ),
                    sm = 12,
                    md = 6,
                    lg = 3,
                ),
                dbc.Col(
                    html.Div(
                        dbc.Select(
                            id = 'subaction-dropdown',
                            options = [],
                            className = 'custom-select input-text'
                        ),
                        id = 'subaction-dropdown-div',
                        className = 'dbc_dark input-text'
                    ),
                    sm = 12,
                    md = 6,
                    lg = 3,
                ),
                dbc.Col(
                    html.Div(
                        dbc.InputGroup(
                            children = [
                                dbc.Button(
                                    'Clear',
                                    color = 'link',
                                    id = 'clear-subaction-dropdown-text-button',
                                    size = 'sm',
                                    style = {'text-decoration': 'none'},
                                ),
                                dbc.Input(
                                    id = 'subaction-dropdown-text',
                                    placeholder = 'Positional arguments',
                                    className = 'input-text',
                                ),
                            ],
                        ),
                    ),
                    id = 'subaction-dropdown-text-div',
                    lg = 6,
                    md = 12,
                    sm = 12,
                    className = 'dbc_dark',
                ),
            ]
        ),
        html.Br(),
        dbc.Row(
            children = [
                dbc.Col(
                    html.Div(
                        dcc.Dropdown(
                            id = 'flags-dropdown',
                            multi = True,
                            placeholder = 'Boolean flags',
                            options = [],
                            value = ['--yes'],
                        ),
                        id = 'flags-dropdown-div',
                        className = 'dbc_dark input-text',
                    ),
                    width = widths['flags'],
                ),
            ],
        ),
        html.Br(),
        html.Div(id='input-flags-div'),
        dbc.Row(
            children = [
                dbc.Col(
                    children = [
                        html.Div(
                            children = [
                                dbc.Button(
                                    'Additional parameters',
                                    id = 'show-arguments-collapse-button',
                                    color = 'link',
                                    size = 'md',
                                    outline = True,
                                    style = {'display': 'none'},
                                ),
                                #  html.Br(),
                                dbc.Collapse(
                                    children = [
                                        dbc.Button(
                                            'Clear',
                                            id = 'clear-begin-end-datepicker-button',
                                            color = 'link',
                                            size = 'sm',
                                        ),
                                        dcc.DatePickerRange(
                                            id = 'begin-end-datepicker',
                                        ),
                                    ],
                                    id = 'arguments-collapse',
                                ),
                            ], ### end of div children
                        ),
                    ], ### end of col children
                    width = widths['arguments'],
                ),
            ], ### end of row children
        ),
    ], ### end of parent div children
    id = 'action-div',
)


dropdown_keys_row = dbc.Row(
    [
        dbc.Col(
            html.Div(
                [
                    dcc.Dropdown(
                        id = 'connector-keys-dropdown',
                        options = [],
                        placeholder = placeholders['ck'],
                        multi = True,
                    ),
                ],
                className = 'dbc_dark',
            ),
            lg = 4,
            md = 12,
            sm = 12,
        ),
        dbc.Col(
            html.Div(
                [
                    dcc.Dropdown(
                        id = 'metric-keys-dropdown',
                        options = [],
                        placeholder = placeholders['mk'],
                        multi = True,
                    ),
                ],
                className = 'dbc_dark'
            ),
            lg = 4,
            md = 12,
            sm = 12,
        ),
        dbc.Col(
            html.Div(
                [
                    dcc.Dropdown(
                        id = 'location-keys-dropdown',
                        options = [],
                        placeholder = placeholders['lk'],
                        multi = True,
                    ),
                ],
                className = 'dbc_dark'
            ),
            lg = 4,
            md = 12,
            sm = 12,
        ),
    ] ### end of filters row children
)
dropdown_tab_content = html.Div([
    dbc.Card(
        dbc.CardBody(
            [
                #  html.P('Pipe Keys'),
                dropdown_keys_row,
            ], ### end of card children
            className = 'card-text',
        )
    ),
    html.Br(),
    dbc.Card(
        dbc.CardBody(
            [
                action_dropdown_row,
            ],
            className = 'card-text',
        ),
    ),
])

text_tab_content = dbc.Card(
    dbc.CardBody(
        [
            dbc.Row(
                [
                    dbc.Col(html.Div(className='dbc_dark', children=[
                        dbc.InputGroup(
                            [
                                #  dbc.InputGroupAddon(
                                    dbc.Button(
                                        'Clear',
                                        id = 'clear-connector-keys-input-button',
                                        color = 'link',
                                        size = 'sm',
                                    ),
                                    #  addon_type = 'prepend',
                                #  ),
                                dbc.Input(
                                    id = 'connector-keys-input',
                                    placeholder = placeholders['ck'],
                                    type = 'text',
                                    value = '',
                                    list = 'connector-keys-list',
                                    className = 'dbc_dark'
                                ),
                            ],
                            size = input_group_sizes['ck'],
                        )]),
                        width = 4,
                    ),
                    dbc.Col(
                        dbc.InputGroup(
                            [
                                #  dbc.InputGroupAddon(
                                    dbc.Button(
                                        'Clear',
                                        id = 'clear-metric-keys-input-button',
                                        color = 'link',
                                        size = 'sm',
                                    ),
                                    #  addon_type = 'prepend',
                                #  ),
                                dbc.Input(
                                    id = 'metric-keys-input',
                                    placeholder = placeholders['mk'],
                                    type = 'text',
                                    value = '',
                                    list = 'metric-keys-list',
                                ),
                            ],
                            size = input_group_sizes['mk'],
                        ),
                        width = 4,
                    ),
                    dbc.Col(
                        dbc.InputGroup(
                            [
                                #  dbc.InputGroupAddon(
                                    dbc.Button(
                                        'Clear',
                                        id = 'clear-location-keys-input-button',
                                        color = 'link',
                                        size = 'sm',
                                    ),
                                    #  addon_type = 'prepend',
                                #  ),
                                dbc.Input(
                                    id = 'location-keys-input',
                                    placeholder = placeholders['lk'],
                                    type = 'text',
                                    value = '',
                                    list = 'location-keys-list',
                                ),
                            ],
                            size = input_group_sizes['lk'],
                        ),
                        width = 4,
                    ),
                ]
            ),
            html.Br(),
            dbc.Row(
                dbc.Col(
                    dbc.InputGroup(
                        [
                            #  dbc.InputGroupAddon(
                                #  dbc.Button(
                                    #  'Clear',
                                    #  id = 'clear-params-textarea-button',
                                    #  color = 'link',
                                    #  size = 'sm',
                                #  ),
                                #  addon_type = 'prepend',
                            #  ),
                            search_parameters_editor,
                            #  dbc.Textarea(
                                #  id = 'params-textarea',
                                #  placeholder = placeholders['params'],
                                #  value = '',
                            #  )
                        ],
                        size = input_group_sizes['params'],
                    )
                )
            ),
        ]
    )
)

keys_lists_content = html.Div([
    html.Datalist(id='connector-keys-list'),
    html.Datalist(id='metric-keys-list'),
    html.Datalist(id='location-keys-list'),
], hidden=True)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for bootstrapping elements
(pipes, configuration, etc)
"""

from __future__ import annotations
from meerschaum.utils.typing import Union, Any, Sequence, SuccessTuple, Optional, Tuple, List

def bootstrap(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Launch an interactive wizard to bootstrap pipes or connectors.
    
    Example:
        `bootstrap pipes`

    """
    from meerschaum.actions import choose_subaction
    options = {
        'pipes'      : _bootstrap_pipes,
        'connectors' : _bootstrap_connectors,
    }
    return choose_subaction(action, options, **kw)


def _bootstrap_pipes(
        action: Optional[List[str]] = None,
        connector_keys: Optional[List[str]] = None,
        metric_keys: Optional[List[str]] = None,
        location_keys: Optional[List[Optional[str]]] = None,
        yes: bool = False,
        force: bool = False,
        noask: bool = False,
        debug: bool = False,
        mrsm_instance: Optional[str] = None,
        shell: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Create a new pipe.
    If no keys are provided, guide the user through the steps required.

    """
    from meerschaum import get_pipes
    from meerschaum.config import get_config
    from meerschaum.utils.warnings import info, warn, error
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.prompt import yes_no, prompt, choose
    from meerschaum.connectors.parse import is_valid_connector_keys, parse_instance_keys
    from meerschaum.utils.misc import get_connector_labels
    from meerschaum.utils.formatting._shell import clear_screen

    instance_connector = parse_instance_keys(mrsm_instance)

    if connector_keys is None:
        connector_keys = []
    if metric_keys is None:
        metric_keys = []
    if location_keys is None:
        location_keys = []

    _clear = get_config('shell', 'clear_screen', patch=True)
    abort_tuple = (False, "No pipes were bootstrapped.")

    if (
        len(connector_keys) > 0 and
        len(metric_keys) > 0
    ):
        info(
            "You've provided the following keys:\n" +
            "  - Connector Keys: " + str(connector_keys) + "\n" +
            "  - Metric Keys: " + str(metric_keys) + "\n" +
            (
                ("  - Location Keys: " + str(location_keys) + "\n")
                if len(location_keys) > 0 else ""
            )
        )
        if not force:
            try:
                if not yes_no(
                    ("Would you like to bootstrap pipes with these keys?\n"
                    + "Existing pipes will be deleted!"),
                    default = 'n',
                    yes = yes,
                    noask = noask
                ):
                    return abort_tuple
            except KeyboardInterrupt:
                return abort_tuple
    else:
        ### Get the connector.
        new_label = 'New'
        info(
            f"To create a pipe without explicitly using a connector, "
            + "use the `register pipes` command.\n"
        )
        try:
            ck = choose(
                f"Where are the data coming from?\n\n" +
                f"    Please type the keys of a connector from below,\n" +
                f"    or enter '{new_label}' to register a new connector.\n\n" +
                f" {get_config('formatting', 'emoji', 'connector')} Connector:\n",
                get_connector_labels() + [new_label],
                numeric = False
            )
        except KeyboardInterrupt:
            return abort_tuple

        if ck == new_label:
            if _clear:
                clear_screen(debug=debug)
            while True:
                tup = _bootstrap_connectors(
                    [], yes=yes, force=force, debug=debug, return_keys=True, **kw
                )
                if isinstance(tup[0], str):
                    if _clear:
                        clear_screen(debug=debug)
                    ck = tup[0] + ':' + tup[1]
                    break
                elif isinstance(tup[0], bool) and not tup[0]:
                    return abort_tuple
                warn(f"Please register a new connector or press CTRL+C to cancel.", stack=False)
        connector_keys = [ck]

        ### Get the metric.
        while True:
            if _clear:
                clear_screen(debug=debug)
            try:
                mk = prompt(
                    f"What kind of data is this?\n\n" +
                    f"    The metric is the label for the contents of the pipe.\n" +
                    f"    For example, 'weather' might be a metric for weather station data.\n\n" +
                    f" {get_config('formatting', 'emoji', 'metric')} Metric:"
                )
            except KeyboardInterrupt:
                return abort_tuple
            if mk:
                break
            warn("Please enter a metric.", stack=False)
        metric_keys = [mk]

        ### Get the location
        if _clear:
            clear_screen(debug=debug)
        try:
            lk = prompt(
                "Where are the data located?\n\n"
                + "    You can build pipes that share a connector and metric\n"
                + "    but are in different locations.\n\n"
                + f"    For example, you could create Pipe('{ck}', '{mk}', 'home') and\n"
                + f"    Pipe('{ck}', '{mk}', 'work'), which would share a connector\n"
                + "    and metric, but may come from different tables.\n\n"
                + "    In most cases, you can omit the location.\n\n"
                + f" {get_config('formatting', 'emoji', 'location')} Location (Empty to omit):"
            )
        except KeyboardInterrupt:
            return abort_tuple
        lk = None if lk == '' else lk
        location_keys = [lk]

    if _clear:
        clear_screen(debug=debug)
    _pipes = get_pipes(
        connector_keys, metric_keys, location_keys,
        method = 'explicit',
        as_list = True,
        debug = debug,
        mrsm_instance = instance_connector,
        **kw
    )
    pipes = []
    for p in _pipes:
        if p.get_id(debug=debug) is not None and not force:
            try:
                if not yes_no(
                    f"{p} already exists.\n\n    Delete {p}?\n    Data will be lost!",
                    default='n'
                ):
                    info(f"Skipping bootstrapping {p}...")
                    continue
            except KeyboardInterrupt:
                return abort_tuple
        pipes.append(p)

    if len(pipes) == 0:
        return abort_tuple

    success_dict = {}
    successes, failures = 0, 0
    for p in pipes:
        try:
            tup = p.bootstrap(
                interactive=True, force=force, noask=noask, yes=yes, shell=shell, debug=debug
            )
        except Exception as e:
            import traceback
            traceback.print_exc()
            tup = False, f"Failed to bootstrap {p} with exception:\n" + str(e)
        success_dict[p] = tup
        if tup[0]:
            successes += 1
        else:
            failures += 1

    msg = (
        f"Finished bootstrapping {len(pipes)} pipe" + ("s" if len(pipes) != 1 else "") + "\n" +
        f"    ({successes} succeeded, {failures} failed)."
    )

    return (successes > 0), msg

def _bootstrap_connectors(
        action : Optional[List[str]] = None,
        connector_keys : Optional[List[str]] = None,
        yes : bool = False,
        force : bool = False,
        noask : bool = False,
        debug : bool = False,
        return_keys : bool = False,
        **kw : Any
    ) -> Union[SuccessTuple, Tuple[str, str]]:
    """
    Prompt the user for the details necessary to create a Connector.

    """
    from meerschaum.connectors.parse import is_valid_connector_keys
    from meerschaum.connectors import connectors, get_connector, types, custom_types
    from meerschaum.utils.prompt import prompt, yes_no, choose
    from meerschaum.config import get_config
    from meerschaum.config._edit import write_config
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.connectors import attributes as connector_attributes
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.misc import is_int

    abort_tuple = False, "No connectors bootstrapped."
    _clear = get_config('shell', 'clear_screen', patch=True)

    if action is None:
        action = []
    if connector_keys is None:
        connector_keys = []

    if len(connector_keys) == 0:
        pass

    try:
        _type = choose(
            (
                'Please choose a connector type.\n'
                + 'For more information on connectors, '
                + 'please visit https://meerschaum.io/reference/connectors'
            ),
            sorted(list(connectors)),
            default = 'sql'
        )
    except KeyboardInterrupt:
        return abort_tuple

    if _clear:
        clear_screen(debug=debug)

    existing_labels = get_config('meerschaum', 'connectors', _type, warn=False) or []
    _label_choices = sorted([
        label
        for label in existing_labels
        if label != 'default' and label is not None
    ])
    new_connector_label = 'New connector'
    _label_choices.append(new_connector_label)
    while True:
        try:
            _label = prompt(f"New label for '{_type}' connector:")
        except KeyboardInterrupt:
            return abort_tuple
        if _label in existing_labels:
            warn(f"Connector '{_type}:{_label}' already exists.", stack=False)
            overwrite = yes_no(
                f"Do you want to overwrite connector '{_type}:{_label}'?",
                default = 'n',
                yes = yes,
                noask = noask,
            )
            if not overwrite and not force:
                return False, f"No changes made to connector configuration."
                break
        elif _label == "":
            warn(f"Please enter a label.", stack=False)
        else:
            break

    cls = types.get(_type)
    cls_required_attrs = getattr(cls, 'REQUIRED_ATTRIBUTES', [])
    type_attributes = connector_attributes.get(_type, {'required': cls_required_attrs})

    new_attributes = {}
    if 'flavors' in type_attributes:
        try:
            flavor = choose(
                f"Flavor for connector '{_type}:{_label}':",
                sorted(list(type_attributes['flavors'])),
                default = (
                    'timescaledb'
                    if 'timescaledb' in type_attributes['flavors']
                    else None
                )
            )
        except KeyboardInterrupt:
            return abort_tuple
        new_attributes['flavor'] = flavor
        required = sorted(list(connector_attributes[_type]['flavors'][flavor]['requirements']))
        default = type_attributes['flavors'][flavor].get('defaults', {})
    else:
        required = sorted(list(type_attributes.get('required', {})))
        default = type_attributes.get('default', {})
    info(
        f"Please answer the following questions to configure the new connector '{_type}:{_label}'."
        + '\n' + "Press [Ctrl + C] to skip."
    )
    for r in required:
        try:
            val = prompt(f"Value for {r}:")
        except KeyboardInterrupt:
            continue
        if is_int(val):
            val = int(val)
        new_attributes[r] = val

    for k, v in default.items():
        ### skip already configured attributes, (e.g. flavor or from required)
        if k in new_attributes:
            continue
        try:
            val = prompt(f"Value for {k}:", default=str(v))
        except KeyboardInterrupt:
            continue
        if is_int(val):
            val = int(val)
        new_attributes[k] = val

    if _clear:
        clear_screen(debug=debug)
    try:
        conn = get_connector(_type, _label, **new_attributes)
    except Exception as e:
        return False, f"Failed to bootstrap connector '{_type}:{_label}' with exception:\n{e}"

    pprint(new_attributes)
    try:
        ok = (
            yes_no(
                f"Are you ok with these new attributes for connector '{conn}'?",
                default = 'y',
                noask = noask,
                yes = yes
            ) if not yes else yes
        )
    except KeyboardInterrupt:
        ok = False
    if not ok:
        return False, "No changes made to connectors configuration."

    meerschaum_config = get_config('meerschaum')

    if 'connectors' not in meerschaum_config:
        meerschaum_config['connectors'] = {}
    if _type not in meerschaum_config['connectors']:
        meerschaum_config['connectors'][_type] = {}
    meerschaum_config['connectors'][_type][_label] = new_attributes
    write_config({'meerschaum': meerschaum_config}, debug=debug)
    if return_keys:
        return _type, _label
    return True, "Success"


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
bootstrap.__doc__ += _choices_docstring('bootstrap')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for fetching new data into the Pipe
"""

from __future__ import annotations
from datetime import timedelta, datetime
import meerschaum as mrsm
from meerschaum.utils.typing import Optional, Any, Union, SuccessTuple, Iterator
from meerschaum.config import get_config
from meerschaum.utils.warnings import warn

def fetch(
        self,
        begin: Union[datetime, str, None] = '',
        end: Optional[datetime] = None,
        check_existing: bool = True,
        sync_chunks: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Union['pd.DataFrame', Iterator['pd.DataFrame'], None]:
    """
    Fetch a Pipe's latest data from its connector.

    Parameters
    ----------
    begin: Union[datetime, str, None], default '':
        If provided, only fetch data newer than or equal to `begin`.

    end: Optional[datetime], default None:
        If provided, only fetch data older than or equal to `end`.

    check_existing: bool, default True
        If `False`, do not apply the backtrack interval.

    sync_chunks: bool, default False
        If `True` and the pipe's connector is of type `'sql'`, begin syncing chunks while fetching
        loads chunks into memory.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A `pd.DataFrame` of the newest unseen data.

    """
    if 'fetch' not in dir(self.connector):
        warn(f"No `fetch()` function defined for connector '{self.connector}'")
        return None

    from meerschaum.connectors import custom_types, get_connector_plugin
    from meerschaum.utils.debug import dprint, _checkpoint

    _chunk_hook = kw.pop('chunk_hook', None)
    kw['workers'] = self.get_num_workers(kw.get('workers', None))
    if sync_chunks and _chunk_hook is None:

        def _chunk_hook(chunk, **_kw) -> SuccessTuple:
            """
            Wrap `Pipe.sync()` with a custom chunk label prepended to the message.
            """
            from meerschaum.config._patch import apply_patch_to_config
            kwargs = apply_patch_to_config(kw, _kw)
            chunk_success, chunk_message = self.sync(chunk, **kwargs)
            chunk_label = self._get_chunk_label(chunk, self.columns.get('datetime', None))
            if chunk_label:
                chunk_message = '\n' + chunk_label + '\n' + chunk_message
            return chunk_success, chunk_message


    with mrsm.Venv(get_connector_plugin(self.connector)):
        df = self.connector.fetch(
            self,
            begin = _determine_begin(
                self,
                begin,
                check_existing = check_existing,
                debug = debug,
            ),
            end = end,
            chunk_hook = _chunk_hook,
            debug = debug,
            **kw
        )
    return df


def get_backtrack_interval(
        self,
        check_existing: bool = True,
        debug: bool = False,
    ) -> Union[timedelta, int]:
    """
    Get the chunk interval to use for this pipe.

    Parameters
    ----------
    check_existing: bool, default True
        If `False`, return a backtrack_interval of 0 minutes.

    Returns
    -------
    The backtrack interval (`timedelta` or `int`) to use with this pipe's `datetime` axis.
    """
    default_backtrack_minutes = get_config('pipes', 'parameters', 'fetch', 'backtrack_minutes')
    configured_backtrack_minutes = self.parameters.get('fetch', {}).get('backtrack_minutes', None)
    backtrack_minutes = (
        configured_backtrack_minutes
        if configured_backtrack_minutes is not None
        else default_backtrack_minutes
    ) if check_existing else 0

    backtrack_interval = timedelta(minutes=backtrack_minutes)
    dt_col = self.columns.get('datetime', None)
    if dt_col is None:
        return backtrack_interval

    dt_dtype = self.dtypes.get(dt_col, 'datetime64[ns]')
    if 'int' in dt_dtype.lower():
        return backtrack_minutes

    return backtrack_interval


def _determine_begin(
        pipe: mrsm.Pipe,
        begin: Union[datetime, int, str] = '',
        check_existing: bool = True,
        debug: bool = False,
    ) -> Union[datetime, int, None]:
    """
    Apply the backtrack interval if `--begin` is not provided.

    Parameters
    ----------
    begin: Union[datetime, int, str], default ''
        The provided begin timestamp.

    check_existing: bool, default True
        If `False`, do not apply the backtrack interval.

    Returns
    -------
    A datetime (or int) value from which to fetch.
    Returns `None` if no begin may be determined.
    """
    if begin != '':
        return begin
    sync_time = pipe.get_sync_time(debug=debug)
    if sync_time is None:
        return sync_time
    backtrack_interval = pipe.get_backtrack_interval(check_existing=check_existing, debug=debug)
    if isinstance(sync_time, datetime) and isinstance(backtrack_interval, int):
        backtrack_interval = timedelta(minutes=backtrack_interval)
    try:
        return sync_time - backtrack_interval
    except Exception as e:
        warn(f"Unable to substract backtrack interval {backtrack_interval} from {sync_time}.")
        return sync_time



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Test SQL utility functions.
"""

import datetime
import pytest
from tests.connectors import conns, get_flavors
from meerschaum.connectors.sql import SQLConnector
from meerschaum.connectors.sql.tools import dateadd_str, table_exists, sql_item_name
import dateutil.parser

@pytest.mark.parametrize("flavor", get_flavors())
def test_dateadd_str(flavor: str):
    """
    Verify that the DATEADD function works for all flavors.
    """
    conn = conns[flavor]
    if conn.type != 'sql':
        return
    td_margin = (
        datetime.timedelta(microseconds=1000)
        if conn.flavor != 'sqlite' else datetime.timedelta(days=1)
    )
    td_advance = datetime.timedelta(days=1)
    dt = datetime.datetime(2022, 1, 2, 3, 4, 5, 678000)
    dt_str = dateadd_str(conn.flavor, begin=dt, number=td_advance.days)
    q = f"SELECT {dt_str}" + ('' if conn.flavor != 'oracle' else ' FROM DUAL')
    dt_val = conn.value(q)
    assert dt_val is not None
    if conn.flavor == 'sqlite':
        dt_val = dateutil.parser.parse(dt_val)
    assert ((dt + td_advance) - dt_val) <= td_margin


@pytest.mark.parametrize("flavor", get_flavors())
def test_exists(flavor: str):
    conn = conns[flavor]
    if conn.type != 'sql':
        return
    tbl = "foo"
    tbl_name = sql_item_name(tbl, conn.flavor)
    conn.exec(f"DROP TABLE {tbl_name}", silent=True)
    assert table_exists(tbl, conn) is False
    assert conn.exec(f"CREATE TABLE {tbl_name} (bar INT)", commit=True, debug=True) is not None
    assert table_exists(tbl, conn, debug=True) is True
    conn.exec(f"DROP TABLE {tbl_name}", silent=True)
    assert table_exists(tbl, conn, debug=True) is False
 

@pytest.mark.parametrize(
    "uri,expected_attributes", [
        (
            'postgresql://mrsm:mrsm@localhost:5432/meerschaum',
            {
                'flavor': 'postgresql',
                'username': 'mrsm',
                'password': 'mrsm',
                'host': 'localhost',
                'port': 5432,
                'database': 'meerschaum',
            }
        ),
        (
            'timescaledb://mrsm:mrsm@localhost:5432/meerschaum',
            {
                'flavor': 'timescaledb',
                'username': 'mrsm',
                'password': 'mrsm',
                'host': 'localhost',
                'port': 5432,
                'database': 'meerschaum',
            }
        ),
        (
            'sqlite:////home/foo/.config/meerschaum/sqlite/mrsm_local.db',
            {
                'flavor': 'sqlite',
                'database': '/home/foo/.config/meerschaum/sqlite/mrsm_local.db',
            }
        ),
        (
            'mssql+pyodbc://sa:supersecureSECRETPASSWORD123!'
            + '@localhost:1439/master?driver=ODBC+Driver+17+for+SQL+Server',
            {
                'flavor': 'mssql',
                'username': 'sa',
                'password': 'supersecureSECRETPASSWORD123!',
                'host': 'localhost',
                'port': 1439,
                'database': 'master',
                'driver': 'ODBC Driver 17 for SQL Server',
            },
        ),
        (
            'http://user:pass@localhost:8000',
            {
                'username': 'user',
                'password': 'pass',
                'host': 'localhost',
                'port': 8000,
                'flavor': 'http',
            }
        ),
    ],
)
def test_parse_uri(uri: str, expected_attributes):
    """
    Text that parsing a URI string returns the expected dictionary.
    """
    assert SQLConnector.parse_uri(uri) == expected_attributes



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Execute Meerschaum Actions via the API
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Union

from meerschaum.api import (
    fastapi, app, endpoints, get_api_connector, debug, manager, private, no_auth
)
from meerschaum.actions import actions
import meerschaum.core
actions_endpoint = endpoints['actions']

@app.get(actions_endpoint, tags=['Actions'])
def get_actions(
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ) -> list:
    return list(actions)

@app.post(actions_endpoint + "/{action}", tags=['Actions'])
def do_action(
        action: str,
        keywords: dict = fastapi.Body(...),
        curr_user = (
            fastapi.Depends(manager) if not no_auth else None
        ),
    ) -> SuccessTuple:
    """
    Perform a Meerschaum action (if permissions allow it).

    Parameters
    ----------
    action: str :
        The action to perform.
        
    keywords: dict :
        The keywords dictionary to pass to the action.

    Returns
    -------
    A `SuccessTuple` of success, message.

    """
    if curr_user is not None and curr_user.type != 'admin':
        from meerschaum.config import get_config
        allow_non_admin = get_config(
            'system', 'api', 'permissions', 'actions', 'non_admin', patch=True
        )
        if not allow_non_admin:
            return False, (
                "The administrator for this server has not allowed users to perform actions.\n\n"
                + "Please contact the system administrator, or if you are running this server, "
                + "open the configuration file with `edit config system` "
                + "and search for 'permissions'. "
                + "\nUnder the keys 'api:permissions:actions', "
                + "you can allow non-admin users to perform actions."
        )

    if action not in actions:
        return False, f"Invalid action '{action}'."
    keywords['mrsm_instance'] = keywords.get('mrsm_instance', str(get_api_connector()))
    _debug = keywords.get('debug', debug)
    keywords.pop('debug', None)
    return actions[action](debug=_debug, **keywords)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This module creates the argparse Parser.
"""

from __future__ import annotations
import sys
import argparse
import json
from datetime import datetime, timedelta, timezone
from meerschaum.utils.typing import Union, Dict, List, Any, Tuple, Callable
from meerschaum.utils.misc import string_to_dict
from meerschaum.utils.warnings import error

class InvalidArgumentException(Exception):
    """
    Raised when argparse failes to parse arguments.
    """

_original_argparse_error = argparse.ArgumentParser.error
_original_argparse_parse_known_args = argparse.ArgumentParser.parse_known_args
def _new_argparse_error(message: str):
    error(message, InvalidArgumentException, stack=False)

class ArgumentParser(argparse.ArgumentParser):
    """Override the built-in `argparse` error handling."""

    def parse_known_args(self, *args, exit_on_error: bool = False, **kw):
        _error_bkp = self.error
        if not exit_on_error:
            self.error = _new_argparse_error
        result = _original_argparse_parse_known_args(self, *args, **kw)
        self.error = _error_bkp
        return result


def parse_datetime(dt_str: str) -> datetime:
    """Parse a string into a datetime."""
    from meerschaum.utils.misc import is_int
    if is_int(dt_str):
        return int(dt_str)

    from meerschaum.utils.packages import attempt_import
    dateutil_parser = attempt_import('dateutil.parser')

    try:
        if dt_str.lower() == 'now':
            dt = datetime.now(timezone.utc).replace(tzinfo=None)
        else:
            dt = dateutil_parser.parse(dt_str)
    except Exception as e:
        dt = None
    if dt is None:
        from meerschaum.utils.warnings import warn, error
        error(f"'{dt_str}' is not a valid datetime format.", stack=False)
    return dt

def parse_help(sysargs : Union[List[str], Dict[str, Any]]) -> None:
    """Parse the `--help` flag to determine which help message to print."""
    from meerschaum._internal.arguments._parse_arguments import parse_arguments, parse_line
    from meerschaum.actions import actions, get_subactions
    import importlib, inspect, textwrap
    from copy import deepcopy
    if isinstance(sysargs, list):
        args = parse_arguments(sysargs)
    elif isinstance(sysargs, dict):
        args = sysargs
    elif isinstance(sysargs, str):
        args = parse_line(sysargs)
    _args = deepcopy(args)
    del _args['action']
    if len(args['action']) == 0:
        return actions['show'](['help'], **_args)
    if args['action'][0] not in actions:
        return actions['show'](['actions'], **_args)

    ### Check for subactions.
    if len(args['action']) > 1:
        try:
            subaction = get_subactions(args['action'][0])[args['action'][1]]
        except Exception as e:
            subaction = None
        if subaction is not None:
            return print(textwrap.dedent(subaction.__doc__))

    try:
        doc = actions[args['action'][0]].__doc__
    except Exception as e:
        doc = None
    if doc is None:
        doc = "No help available for '" + f"{args['action'][0]}" + "'."
    return print(textwrap.dedent(doc))

def parse_version(sysargs : List[str]):
    """Print the Meerschaum version."""
    from meerschaum.config import __version__ as version
    from meerschaum.config import __doc__ as doc
    if '--nopretty' in sysargs:
        return print(version)
    return print(doc)

def get_arguments_triggers() -> Dict[str, Tuple[str]]:
    """ """
    triggers = {}
    _actions = parser._actions
    for _a in _actions:
        triggers[_a.dest] = tuple(_a.option_strings)
    return triggers


parser = ArgumentParser(
    prog = 'mrsm',
    description = "Create and Build Pipes with Meerschaum.",
    usage = "mrsm [actions] {options}",
    add_help = False,
)
_seen_plugin_args = {}

groups = {}
groups['actions'] = parser.add_argument_group(title='Actions options')
groups['pipes'] = parser.add_argument_group(title='Pipes options')
groups['sync'] = parser.add_argument_group(title='Sync options')
groups['api'] = parser.add_argument_group(title='API options')
groups['plugins'] = parser.add_argument_group(title='Plugins options')
groups['packages'] = parser.add_argument_group(title='Packages options')
groups['debug'] = parser.add_argument_group(title='Debugging options')
groups['misc'] = parser.add_argument_group(title='Miscellaneous options')


### Actions options
groups['actions'].add_argument(
    'action', nargs='*', help="Actions list to execute. E.g. `start api`"
)
groups['actions'].add_argument(
    '-h', '--help', action='store_true', help="Print a help message for an action."
)
groups['actions'].add_argument(
    '--loop',
    action="store_true", help="Loop the specified action forever (only for select actions)"
)
groups['actions'].add_argument(
    '-y', '--yes', action="store_true", help="Agree to the default choices for prompts"
)
groups['actions'].add_argument(
    '-f', '--force', action="store_true", help="Override safety checks"
)
groups['actions'].add_argument(
    '--noask', '--no-ask', action='store_true',
    help="Automatically choose the defaults answers to questions. Does not result in data loss.",
)
groups['actions'].add_argument(
    '-d', '--daemon', action='store_true',
    help = "Run an action as a background daemon."
)

groups['actions'].add_argument(
    '--rm', action='store_true', help="Delete a job once it has finished executing."
)
groups['actions'].add_argument(
    '--name', '--job-name', type=str, help=(
        "Assign a name to a job. If no name is provided, a random name will be assigned."
    ),
)
groups['actions'].add_argument(
    '-s', '--schedule', '--cron', type=str,
    help = (
        "Continue executing the action according to a schedule (e.g. 'every 1 seconds'). \n"
        + "Often used with `-d`. See this page for scheduling syntax:\n "
        + "https://red-engine.readthedocs.io/en/stable/condition_syntax/index.html"
    )
)
groups['actions'].add_argument(
    '-A', '--sub-args', nargs=argparse.REMAINDER,
    help = (
        "Provide a list of arguments for subprocesses. " +
        "You can also type sub-arguments in [] instead." +
        " E.g. `stack -A='--version'`, `ls [-lh]`, `echo -A these are sub-arguments`"
    )
)

### Pipes options
groups['pipes'].add_argument(
    '-c', '-C', '--connector-keys', nargs='+',
    help="List of connector keys to search for. e.g. -c sql:main api:main"
)
groups['pipes'].add_argument(
    '-m', '-M', '--metric-keys', nargs='+', help="List of metric keys to include in the search"
)
groups['pipes'].add_argument(
    '-l', '-L', '--location-keys', nargs='+', help="List of location keys to include in the search"
)
groups['pipes'].add_argument(
    '-i', '-I', '--mrsm-instance', '--instance', help=(
        "Connector Keys (type:label) to the Meerschaum instance for Pipe objects. " +
        "If label is omitted, use the configured default interface (usually 'sql:main')"
    )
)
groups['pipes'].add_argument(
    '-t', '--tags', nargs='+', help="Only include pipes with these tags.",
)


### Sync options
groups['sync'].add_argument(
    '--min-seconds', '--cooldown', type=float, help=(
        "The minimum number of seconds between syncing laps. Defaults to 1."
    )
)
groups['sync'].add_argument(
    '--timeout-seconds', '--timeout', type=float,
    help="The maximum number of seconds before cancelling a pipe's syncing job. Defaults to 300."
)
groups['sync'].add_argument(
    '--unblock', action="store_true", help="Run the action asynchronously, if possible.",
)
groups['sync'].add_argument(
    '--async', action="store_true",
    help="Run the action asynchronously, if possible. Alias for --unblock",
)
groups['sync'].add_argument(
    '--begin', type=parse_datetime, help="Specify a begin datetime for syncing or displaying data."
)
groups['sync'].add_argument(
    '--end', type=parse_datetime, help="Specify an end datetime for syncing or displaying data."
)
groups['sync'].add_argument(
    '--chunksize', type=int, help=(
        "Specify the database chunksize. Defaults to 10,000."
    ),
)
groups['sync'].add_argument(
    '--chunk-minutes', type=int, help=(
        "Specify the chunk interval in minutes for verification syncs. Defaults to 1440."
    ),
)
groups['sync'].add_argument(
    '--chunk-hours', type=int, help=(
        "Specify the chunk interval in days for verification syncs. See `--chunk-minutes`."
    ),
)
groups['sync'].add_argument(
    '--chunk-days', type=int, help=(
        "Specify the chunk interval in days for verification syncs. See `--chunk-minutes`."
    ),
)
groups['sync'].add_argument(
    '--verify', action="store_true", help="Completely resync a pipe's entire interval.",
)
groups['sync'].add_argument(
    '--deduplicate', '--dedup', action="store_true",
    help="Remove duplicate rows from a pipe's table.",
)
groups['sync'].add_argument(
    '--bounded', '--bound', action="store_true",
    help = (
        "When verifying, do not sync outside "
        + "the existing oldest and newest datetime bounds."
    )
)
groups['sync'].add_argument(
    '--skip-check-existing', '--allow-duplicates', action='store_true',
    help = (
        "Skip checking for duplicate rows when syncing. " +
        "This drastically improves performance when all rows to be synced are unique. " +
        "For example, this setting is highly recommended for use with IoT devices."
    )
)
groups['sync'].add_argument(
    '--cache', action='store_true',
    help = (
        "When syncing or viewing a pipe's data, sync to a local database for later analysis."
    )
)

### API options
groups['api'].add_argument(
    '-p', '--port', type=int, help="The port on which to run the Web API server"
)
groups['api'].add_argument(
    '--host', type=str,
    help="The host address to bind to for the API server. Defaults to '0.0.0.0'."
)
groups['api'].add_argument(
    '-w', '--workers', type=int,
    help = "How many workers to run a concurrent action (e.g. running the API or syncing pipes)"
)
groups['api'].add_argument(
    '--no-dash', '--nodash', action='store_true',
    help = 'When starting the API, do not start the Web interface.',
)
groups['api'].add_argument(
    '--private', '--private-mode', action='store_true',
    help = 'Require authentication for all endpoints.',
)
groups['api'].add_argument(
    '--secure', action='store_true',
    help = "Start the API in secure mode, only allowing `admin` users to execute actions.",
)
groups['api'].add_argument(
    '--no-auth', '--noauth', '--insecure', action='store_true',
    help = 'When starting the API, do not require authentication. WARNING: This is dangerous!',
)
groups['api'].add_argument(
    '--production', '--gunicorn', action='store_true',
    help = 'Start the API with the Gunicorn process manager.'
)
groups['api'].add_argument(
    '--keyfile', type=str,
    help = "Start the API server with this keyfile (requires --certfile).",
)
groups['api'].add_argument(
    '--certfile', type=str,
    help = "Start the API server with this certfile (requires --keyfile).",
)
### Plugins options
groups['plugins'].add_argument(
    '-r', '--repository', '--repo', type=str,
    help="Meerschaum plugins repository to connect to. Specify an API label (default: 'mrsm')"
)

### Packages options
groups['packages'].add_argument(
    '--venv', type=str,
    help="Choose which virtual environment to target when installing or removing packages.",
)
### Debugging Arguments
groups['debug'].add_argument(
    '--debug', action="store_true", help="Print debug statements (max verbosity)"
)
groups['debug'].add_argument(
    '--trace', action="store_true",
    help="Trace execution and open browser windows to visualize the stack."
)

### Miscellaneous arguments
groups['misc'].add_argument(
    '-V', '--version', action="store_true",
    help="Print the Meerschaum version and exit. Has no effect from within the shell."
)
groups['misc'].add_argument(
    '--patch', action="store_true", help="Patch parameters instead of overwriting."
)
groups['misc'].add_argument(
    '--nopretty', action="store_true", help="Print elements without 'pretty' formatting"
)
groups['misc'].add_argument(
    '--skip-deps', action="store_true", help="Skip dependencies when installing plugins.",
)
groups['misc'].add_argument(
    '-P', '--params', type=string_to_dict, help=(
        "Parameters dictionary in JSON format or simple format. " +
        "Simple format is for one-depth dictionaries and does not need braces or quotes." +
        "\nJSON Example:\n" +
        "--params '{\"meerschaum\" : { \"connectors\" : " +
        "\"main\" : { \"host\" : \"localhost\" } } }'" +
        "\n\nSimple example:\n" +
        "--params key1:value1,key2:value2"
    )
)
groups['misc'].add_argument(
    '--temporary', '--temp',
    action = 'store_true',
    help = (
        "Skip creating or modifying instance tables when working with pipes "
        + "(plugins and users still trigger table creation)."
    ),
)
groups['misc'].add_argument(
    '--gui', action='store_true',
    help="Open a DataFrame in an interactive pandasgui or matplotlib window."
)
groups['misc'].add_argument(
    '--shell', action='store_true',
    help="Open the Meerschaum shell and execute the provided action."
)
groups['misc'].add_argument(
    '--use-bash', action='store_true',
    help="Execute non-implemented actions via `bash -c`. Default behavior is to execute directly."
)
groups['misc'].add_argument(
    '--allow-shell-job', action='store_true',
    help = (
        "Allow shell commands to be run as background jobs. "
        + "Default behavior is to only allow recognized actions."
    )
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
This module contains functions for printing elements.
"""

from __future__ import annotations
import meerschaum as mrsm
from meerschaum.utils.typing import SuccessTuple, Union, Sequence, Any, Optional, List, Dict, Tuple

def show(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """Show elements of a certain type.
    
    Command:
        `show {option}`
    
    Example:
        `show pipes`
    """
    from meerschaum.actions import choose_subaction
    show_options = {
        'actions'    : _show_actions,
        'pipes'      : _show_pipes,
        'config'     : _show_config,
        'environment': _show_environment,
        'version'    : _show_version,
        'connectors' : _show_connectors,
        'arguments'  : _show_arguments,
        'data'       : _show_data,
        'columns'    : _show_columns,
        'rowcounts'  : _show_rowcounts,
        'plugins'    : _show_plugins,
        'packages'   : _show_packages,
        'help'       : _show_help,
        'users'      : _show_users,
        'jobs'       : _show_jobs,
        'logs'       : _show_logs,
        'tags'       : _show_tags,
    }
    return choose_subaction(action, show_options, **kw)


def _complete_show(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """
    from meerschaum.actions.start import _complete_start_jobs

    if action is None:
        action = []

    options = {
        'connector' : _complete_show_connectors,
        'connectors': _complete_show_connectors,
        'config'    : _complete_show_config,
        'package'   : _complete_show_packages,
        'packages'  : _complete_show_packages,
        'job'       : _complete_start_jobs,
        'jobs'      : _complete_start_jobs,
        'log'       : _complete_start_jobs,
        'logs'      : _complete_start_jobs,
    }

    if (
        len(action) > 0 and action[0] in options
            and kw.get('line', '').split(' ')[-1] != action[0]
    ):
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum._internal.shell import default_action_completer
    return default_action_completer(action=(['show'] + action), **kw)


def _show_actions(**kw: Any) -> SuccessTuple:
    """
    Show available actions.
    """
    from meerschaum.actions import actions
    from meerschaum.utils.misc import print_options
    from meerschaum._internal.shell.Shell import hidden_commands
    _actions = [ _a for _a in actions if _a not in hidden_commands ]
    print_options(options=_actions, name='actions', actions=False, **kw)
    return True, "Success"


def _show_help(**kw: Any) -> SuccessTuple:
    """
    Print the --help menu from `argparse`.
    """
    from meerschaum._internal.arguments._parser import parser
    print(parser.format_help())
    return True, "Success"


def _show_config(
        action: Optional[List[str]] = None,
        debug: bool = False,
        nopretty: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the configuration dictionary.
    Sub-actions defined in the action list are recursive indices in the config dictionary.
    
    Example:
        `show config pipes` -> cf['pipes']
    """
    import json
    from meerschaum.utils.formatting import pprint
    from meerschaum.config import get_config
    from meerschaum.config._paths import CONFIG_DIR_PATH
    from meerschaum.utils.debug import dprint

    if action is None:
        action = []

    valid, config = get_config(*action, as_tuple=True, warn=False)
    if not valid:
        return False, f"Invalid configuration keys '{action}'."
    if nopretty:
        print(json.dumps(config))
    else:
        pprint(config)
    return (True, "Success")


def _complete_show_config(action: Optional[List[str]] = None, **kw : Any):
    from meerschaum.config._read_config import get_possible_keys
    keys = get_possible_keys()
    if not action:
        return keys
    possibilities = []
    for key in keys:
        if key.startswith(action[0]) and action[0] != key:
            possibilities.append(key)
    return possibilities


def _show_pipes(
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Print a stylized tree of available Meerschaum pipes.
    Respects global ANSI and UNICODE settings.
    """
    import json
    from meerschaum import get_pipes
    from meerschaum.utils.misc import flatten_pipes_dict
    from meerschaum.utils.formatting import ANSI, pprint_pipes
    pipes = get_pipes(debug=debug, **kw)

    if len(pipes) == 0:
        return False, "No pipes to show."

    if len(flatten_pipes_dict(pipes)) == 1:
        return flatten_pipes_dict(pipes)[0].show(debug=debug, nopretty=nopretty, **kw)

    if not nopretty:
        pprint_pipes(pipes)
    else:
        pipes_list = flatten_pipes_dict(pipes)
        for p in pipes_list:
            print(json.dumps(p.__getstate__()))

    return (True, "Success")


def _show_version(nopretty: bool = False, **kw : Any) -> SuccessTuple:
    """
    Show the Meerschaum doc string.

    Examples:
        - `show version`
        - `show version --nopretty`
    """
    from meerschaum import __version__ as version
    _print = print
    if nopretty:
        msg = version
    else:
        from meerschaum.utils.warnings import info
        msg = "Meerschaum v" + version
        _print = info
    _print(msg)
    return (True, "Success")


def _show_connectors(
        action: Optional[List[str]] = None,
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show connectors configuration and, if requested, specific connector attributes.
    
    Examples:
        `show connectors`
            Display the connectors configuration.
    
        `show connectors sql:main`
            Show the connectors configuration and the attributes for the connector 'sql:main'.
    """
    from meerschaum.connectors import connectors
    from meerschaum.config import get_config
    from meerschaum.utils.formatting import make_header
    from meerschaum.utils.formatting import pprint
    if not nopretty:
        print(make_header("\nConfigured connectors:"))
    pprint(get_config('meerschaum', 'connectors'), nopretty=nopretty)
    if not nopretty:
        print(make_header("\nActive connectors:"))
        pprint(connectors, nopretty=nopretty)

    from meerschaum.connectors.parse import parse_instance_keys
    if action:
        attr, keys = parse_instance_keys(action[0], construct=False, as_tuple=True, debug=debug)
        if attr:
            if not nopretty:
                print(make_header("\n" + f"Attributes for connector '{keys}':"))
            pprint(attr, nopretty=nopretty)

    return True, "Success"


def _complete_show_connectors(
        action: Optional[List[str]] = None, **kw: Any
    ) -> List[str]:
    from meerschaum.utils.misc import get_connector_labels
    _text = action[0] if action else ""
    return get_connector_labels(search_term=_text, ignore_exact_match=True)


def _show_arguments(
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the parsed keyword arguments.
    """
    from meerschaum.utils.formatting import pprint
    pprint(kw)
    return True, "Success"


def _show_data(
        action: Optional[List[str]] = None,
        gui: bool = False,
        begin: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        params: Optional[Dict[str, Any]] = None,
        chunksize: Optional[int] = -1,
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show pipes data as Pandas DataFrames.
    
    Usage:
        - Use --gui to open an interactive window.
    
        - `show data all` to grab all data for the chosen Pipes.
          WARNING: This may be dangerous!
    
        - `show data 60` to grab the last 60 (or any number) minutes of data for all pipes.
    
        - `show data --begin 2020-01-01 --end 2021-01-01` to specify date rangers.
          **NOTE:** You must specify to at least the day, otherwise the date parser will assume
          you mean today's date.
    
        - Regular pipes parameters (-c, -m, -l, etc.)
    
    Examples:
        - show data -m weather --gui
            Open an interactive pandasgui window for the last 1440 minutes of data
            for all pipes of metric 'weather'.
    """
    import sys, json
    from meerschaum import get_pipes
    from meerschaum.utils.packages import attempt_import, import_pandas
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.formatting import pprint
    pd = import_pandas()

    if action is None:
        action = []

    pipes = get_pipes(as_list=True, params=params, debug=debug, **kw)
    try:
        backtrack_minutes = float(action[0])
    except Exception as e:
        backtrack_minutes = (
            1440 if (
                begin is None
                and end is None
                and (
                    not action
                    or (action and action[0] != 'all')
                )
            ) else None
        )

    for p in pipes:
        try:
            if backtrack_minutes is not None:
                df = p.get_backtrack_data(
                    backtrack_minutes = backtrack_minutes,
                    chunksize = chunksize,
                    params = params,
                    debug = debug,
                )
            else:
                df = p.get_data(
                    begin = begin,
                    end = end,
                    params = params,
                    chunksize = chunksize,
                    debug = debug,
                )
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            df = None

        if df is None:
            warn(f"Failed to fetch data for {p}.", stack=False)
            continue

        info_msg = (
            (
                f"Last {backtrack_minutes} minute"
                + ('s' if backtrack_minutes != 1 else '')
                + f" of data for {p}:"
            ) if backtrack_minutes is not None
            else (
                f"Data for {p}" +
                    (f" from {begin}" if begin is not None else '') +
                    (f" to {end}" if end is not None else '') + ':'
            )
        )

        if not nopretty:
            info(info_msg)
        else:
            print(json.dumps(p.__getstate__()))
            df = df.fillna(pd.NA).to_json(orient='columns')

        pprint(df, nopretty=nopretty)

    return True, "Success"


def _show_columns(
        action: Optional[List[str]] = None,
        debug: bool = False,
        nopretty: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the registered and table columns for pipes.
    """
    from meerschaum import get_pipes
    from meerschaum.utils.formatting._pipes import pprint_pipe_columns
    pipes = get_pipes(as_list=True, debug=debug, **kw)
    for p in pipes:
        pprint_pipe_columns(p, nopretty=nopretty, debug=debug)

    return True, "Success"


def _show_rowcounts(
        action: Optional[List[str]] = None,
        workers: Optional[int] = None,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the rowcounts for pipes.
    
    To see remote rowcounts (execute `COUNT(*)` on the source server),
    execute `show rowcounts remote`.
    """
    from meerschaum.utils.misc import print_options
    from meerschaum.utils.pool import get_pool
    from meerschaum import get_pipes

    if action is None:
        action = []
    remote = action and action[0] == 'remote'

    pipes = get_pipes(as_list=True, debug=debug, **kw)
    pool = get_pool(workers=workers)
    def _get_rc(_pipe):
        return _pipe.get_rowcount(remote=remote, debug=debug)

    rowcounts = pool.map(_get_rc, pipes) if pool is not None else [_get_rc(p) for p in pipes]

    rc_dict = {}
    for i, p in enumerate(pipes):
        rc_dict[p] = rowcounts[i]

    msgs = []
    for p, rc in rc_dict.items():
        msgs.append(f'{p}\n{rc}\n')

    header = "Remote row-counts:" if remote else "Pipe row-counts:"

    print_options(msgs, header=header, **kw)

    return True, "Success"

def _show_plugins(
        action: Optional[List[str]] = None,
        repository: Optional[str] = None,
        nopretty: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the installed plugins.
    """
    from meerschaum.plugins import import_plugins, get_plugins_names
    from meerschaum.utils.misc import print_options
    from meerschaum.connectors.parse import parse_repo_keys
    from meerschaum.utils.warnings import info
    from meerschaum.core import User
    repo_connector = parse_repo_keys(repository)

    if action is None:
        action = []

    if action == [''] or len(action) == 0:
        _to_print = get_plugins_names()
        header = "Installed plugins:"
        if not nopretty:
            info(
                f"To see all installable plugins from repository '{repo_connector}', "
                + "run `show plugins all`"
            )
            info("To see plugins created by a certain user, run `show plugins [username]`")
    elif action[0] in ('all'):
        _to_print = repo_connector.get_plugins(debug=debug)
        header = f"Available plugins from Meerschaum repository '{repo_connector}':"
    else:
        username = action[0]
        user_id = repo_connector.get_user_id(User(username, ''))
        _to_print = repo_connector.get_plugins(user_id=user_id, debug=debug)
        header = f"Plugins from user '{username}' at Meerschaum repository '{repo_connector}':"

    print_options(_to_print, header=header, debug=debug, nopretty=nopretty, **kw)

    return True, "Success"

def _show_users(
        mrsm_instance: Optional[str] = None,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the registered users in a Meerschaum instance (default is the current instance).
    """
    from meerschaum.config import get_config
    from meerschaum.connectors.parse import parse_instance_keys
    from meerschaum.utils.misc import print_options
    instance_connector = parse_instance_keys(mrsm_instance)
    users_list = instance_connector.get_users(debug=debug)

    try:
        users_list = instance_connector.get_users(debug=debug)
    except Exception as e:
        return False, f"Failed to get users from instance '{mrsm_instance}'"

    print_options(users_list, header=f"Registered users for instance '{instance_connector}':")

    return True, "Success"

def _show_packages(
        action: Optional[List[str]] = None,
        nopretty: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the packages in dependency groups, or as a list with `--nopretty`.
    """
    from meerschaum.utils.packages import packages
    from meerschaum.utils.warnings import warn

    if action is None:
        action = []

    if not nopretty:
        from meerschaum.utils.formatting import pprint

    def _print_packages(_packages):
        for import_name, install_name in _packages.items():
            print(install_name)

    _print_func = pprint if not nopretty else _print_packages

    key = 'full' if len(action) == 0 else action[0]

    try:
        _print_func(packages[key])
    except KeyError:
        warn(f"'{key}' is not a valid group.", stack=False)

    return True, "Success"

def _complete_show_packages(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    from meerschaum.utils.packages import packages
    if not action:
        return sorted(list(packages.keys()))
    possibilities = []

    for key in packages:
        if key.startswith(action[0]) and action[0] != key:
            possibilities.append(key)

    return possibilities

def _show_jobs(
        action: Optional[List[str]] = None,
        nopretty: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Show the currently running and stopped jobs.
    """
    from meerschaum.utils.daemon import get_filtered_daemons
    from meerschaum.utils.formatting._jobs import pprint_jobs
    daemons = get_filtered_daemons(action)
    if not daemons:
        if not action and not nopretty:
            from meerschaum.utils.warnings import info
            info('No running or stopped jobs.')
            print(
                f"    You can start a background job with `-d` or `--daemon`,\n" +
                "    or run the command `start job` before action commands.\n\n" +
                "    Examples:\n" +
                "      - start api -d\n" +
                "      - start job sync pipes --loop"
            )
        return False, "No jobs to show."
    pprint_jobs(daemons, nopretty=nopretty)
    return True, "Success"


def _show_logs(
        action: Optional[List[str]] = None,
        nopretty: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Show the logs for jobs.
    
    You may specify jobs to only print certain logs.
    To print the entire log file, add the `--nopretty` flag.
    
    Usage:
        `show logs`
        `show logs --nopretty`
        `show logs myjob myotherjob`
    """
    import os, pathlib, random, asyncio
    from meerschaum.utils.packages import attempt_import, import_rich
    from meerschaum.utils.daemon import get_filtered_daemons, Daemon
    from meerschaum.utils.warnings import warn, info
    from meerschaum.utils.formatting import get_console, ANSI, UNICODE
    from meerschaum.utils.misc import tail
    from meerschaum.config._paths import LOGS_RESOURCES_PATH
    from meerschaum.config import get_config
    if not ANSI:
        info = print
    colors = get_config('jobs', 'logs', 'colors')
    daemons = get_filtered_daemons(action)

    def _build_buffer_spaces(daemons) -> Dict[str, str]:
        _max_len_id = max([len(d.daemon_id) for d in daemons]) if daemons else 0
        _buffer_len = max(get_config('jobs', 'logs', 'min_buffer_len'), _max_len_id + 2)
        return {
            d.daemon_id: ''.join([' ' for i in range(_buffer_len - len(d.daemon_id))])
            for d in daemons
        }

    def _build_job_colors(daemons, _old_job_colors = None) -> Dict[str, str]:
        return {d.daemon_id: colors[i % len(colors)] for i, d in enumerate(daemons)}

    _buffer_spaces = _build_buffer_spaces(daemons)
    _job_colors = _build_job_colors(daemons)

    def _get_buffer_spaces(daemon_id):
        nonlocal _buffer_spaces, daemons
        if daemon_id not in _buffer_spaces:
            d = Daemon(daemon_id=daemon_id)
            if d not in daemons:
                daemons = get_filtered_daemons(action)
            _buffer_spaces = _build_buffer_spaces(daemons)
        return _buffer_spaces[daemon_id]

    def _get_job_colors(daemon_id):
        nonlocal _job_colors, daemons
        if daemon_id not in _job_colors:
            d = Daemon(daemon_id=daemon_id)
            if d not in daemons:
                daemons = get_filtered_daemons(action)
            _job_colors = _build_job_colors(daemons)
        return _job_colors[daemon_id]

    def _follow_pretty_print():
        watchgod = attempt_import('watchgod')
        rich = import_rich()
        rich_text = attempt_import('rich.text')
        _watch_daemon_ids = {d.daemon_id: d for d in daemons}
        info("Watching log files...")

        def _print_job_line(daemon, line):
            text = rich_text.Text(daemon.daemon_id)
            text.append(
                _get_buffer_spaces(daemon.daemon_id) + '| '
                + (line[:-1] if line[-1] == '\n' else line)
            )
            if ANSI:
                text.stylize(
                    _get_job_colors(daemon.daemon_id),
                    0,
                    len(daemon.daemon_id) + len(_get_buffer_spaces(daemon.daemon_id)) + 1
                )
            get_console().print(text)


        def _print_log_lines(daemon):
            for line in daemon.readlines():
                _print_job_line(daemon, line)

        def _seek_back_offset(d) -> bool:
            if d.log_offset_path.exists():
                d.log_offset_path.unlink()

            latest_subfile_path = d.rotating_log.get_latest_subfile_path()
            latest_subfile_index = d.rotating_log.get_latest_subfile_index()
            with open(latest_subfile_path, 'r', encoding='utf-8') as f:
                latest_lines = f.readlines()

            lines_to_show = get_config('jobs', 'logs', 'lines_to_show')
            positions_to_rewind = len(''.join(latest_lines[(-1 * lines_to_show):]))
            backup_index = len(''.join(latest_lines)) - positions_to_rewind

            d.rotating_log._cursor = (
                latest_subfile_index,
                max(0, backup_index)
            )
            d._write_log_offset()
            return True

        daemons_being_watched = {
            d.daemon_id: d
            for d in daemons
        }
        for d in daemons:
            _seek_back_offset(d)
            _print_log_lines(d)

        _quit = False
        async def _watch_logs():
            async for changes in watchgod.awatch(LOGS_RESOURCES_PATH):
                if _quit:
                    return
                for change in changes:
                    file_path_str = change[1]
                    if '.log' not in file_path_str or '.log.offset' in file_path_str:
                        continue
                    file_path = pathlib.Path(file_path_str)
                    if not file_path.exists():
                        continue
                    daemon_id = file_path.name.split('.log')[0]
                    if daemon_id not in _watch_daemon_ids and action:
                        continue
                    try:
                        daemon = Daemon(daemon_id=daemon_id)
                    except Exception as e:
                        daemon = None
                        warn(f"Seeing new logs for non-existent job '{daemon_id}'.", stack=False)

                    if daemon is not None:
                        _print_log_lines(daemon)

        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(_watch_logs())
        except KeyboardInterrupt:
            _quit = True

    def _print_nopretty_log_text():
        for d in daemons:
            log_text = d.log_text
            print(d.daemon_id)
            print(log_text)

    _print_log_text = _follow_pretty_print if not nopretty else _print_nopretty_log_text
    _print_log_text()

    return True, "Success"


def _show_environment(
        nopretty: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Show all of the current environment variables with begin with `'MRSM_'`.
    """
    import os
    from meerschaum.utils.formatting import pprint
    from meerschaum.config._environment import get_env_vars
    pprint(
        {
            env_var: os.environ[env_var]
            for env_var in get_env_vars()
        },
        nopretty = nopretty,
    )
    return True, "Success"


def _show_tags(
        action: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        workers: Optional[int] = None,
        nopretty: bool = False,
        **kwargs
    ) -> SuccessTuple:
    """
    Show the existing tags and their associated pipes.
    """
    import json
    from collections import defaultdict
    import meerschaum as mrsm
    from meerschaum.utils.formatting import pipe_repr, UNICODE, ANSI
    from meerschaum.utils.pool import get_pool
    from meerschaum.config import get_config
    rich_tree, rich_panel, rich_text, rich_console, rich_columns = (
        mrsm.attempt_import('rich.tree', 'rich.panel', 'rich.text', 'rich.console', 'rich.columns')
    )
    panel = rich_panel.Panel.fit('Tags')
    tree = rich_tree.Tree(panel)
    action = action or []
    tags = action + (tags or [])
    pipes = mrsm.get_pipes(as_list=True, tags=tags, **kwargs)
    if not pipes:
        return False, f"No pipes were found with the given tags."

    pool = get_pool(workers=workers)
    tag_prefix = get_config('formatting', 'pipes', 'unicode', 'icons', 'tag') if UNICODE else ''
    tag_style = get_config('formatting', 'pipes', 'ansi', 'styles', 'tags') if ANSI else None

    tags_pipes = defaultdict(lambda: [])
    gather_pipe_tags = lambda pipe: (pipe, (pipe.tags or []))

    pipes_tags = dict(pool.map(gather_pipe_tags, pipes))

    for pipe, tags in pipes_tags.items():
        for tag in tags:
            if action and tag not in action:
                continue
            tags_pipes[tag].append(pipe)

    columns = []
    sorted_tags = sorted([tag for tag in tags_pipes])
    for tag in sorted_tags:
        _pipes = tags_pipes[tag]
        tag_text = (
            rich_text.Text(tag_prefix)
            + rich_text.Text(
                tag,
                style = tag_style,
            )
        )
        pipes_texts = [
            pipe_repr(pipe, as_rich_text=True)
            for pipe in _pipes
        ]
        tag_group = rich_console.Group(*pipes_texts)
        tag_panel = rich_panel.Panel(tag_group, title=tag_text, title_align='left')
        columns.append(tag_panel)

    if len(columns) == 0:
        return False, "No pipes have been tagged."

    if not nopretty:
        mrsm.pprint(
            rich_columns.Columns(
                columns,
                equal = True,
            ),
        )
    else:
        for tag, _pipes in tags_pipes.items():
            print(tag)
            for pipe in _pipes:
                print(json.dumps(pipe.meta))

    return True, "Success"



### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
show.__doc__ += _choices_docstring('show')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Routes for managing plugins
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, SuccessTuple, Union, Any, Dict

from meerschaum.api import (
    fastapi,
    app,
    endpoints,
    get_api_connector,
    pipes,
    get_pipe,
    manager,
    debug,
    private, no_auth,
)
import fastapi
from meerschaum.api.tables import get_tables
from fastapi import FastAPI, File, UploadFile
from meerschaum.utils.packages import attempt_import
import meerschaum.core
from meerschaum.core import Plugin
starlette_responses = attempt_import('starlette.responses', warn=False)
FileResponse = starlette_responses.FileResponse

sqlalchemy = attempt_import('sqlalchemy')
plugins_endpoint = endpoints['plugins']

@app.post(plugins_endpoint + '/{name}', tags=['Plugins'])
def register_plugin(
        name: str,
        version: str = None,
        attributes: str = None,
        archive: UploadFile = File(...),
        curr_user = (
            fastapi.Depends(manager) if not no_auth else None
        ),
    ) -> SuccessTuple:
    """
    Register a plugin and save its archive file.

    Parameters
    ----------
    name: str :
        The name of the plugin.
        
    version: str, default None
        The version of the plugin.

    attributes: str, default None
        JSON-encoded string of the attributes dictionary.

    archive: UploadFile :
        The archive file of the plugin.

    curr_user: 'meerschaum.core.User'
        The logged-in user.

    Returns
    -------
    A `SuccessTuple` of success, message.

    """
    from meerschaum.config import get_config
    allow_plugins = get_config('system', 'api', 'permissions', 'registration', 'plugins')
    if not allow_plugins:
        return False, (
            "The administrator for this server has not allowed plugin registration.\n\n" +
            "Please contact the system administrator, or if you are running this server, " +
            "open the configuration file with `edit config system` and search for 'permissions'. " +
            " Under the keys `api:permissions:registration`, " +
            "you can toggle various registration types."
        )

    import json, shutil, pathlib, os
    get_tables()
    if attributes is None:
        attributes = json.dumps({})
    attributes = json.loads(attributes)
    if isinstance(attributes, str) and attributes[0] == '{':
        try:
            attributes = json.loads(attributes)
        except Exception as e:
            pass

    plugin = Plugin(name, version=version, attributes=attributes)
    if curr_user is not None:
        plugin_user_id = get_api_connector().get_plugin_user_id(plugin)
        curr_user_id = get_api_connector().get_user_id(curr_user) if curr_user is not None else -1
        if plugin_user_id is not None and plugin_user_id != curr_user_id:
            return False, f"User '{curr_user.username}' cannot edit plugin '{plugin}'."
        plugin.user_id = curr_user_id
    else:
        plugin.user_id = -1

    success, msg = get_api_connector().register_plugin(plugin, make_archive=False, debug=debug)

    ### TODO delete and install new version of plugin on success
    if success:
        archive_path = plugin.archive_path
        temp_archive_path = pathlib.Path(str(archive_path) + '.tmp')
        with temp_archive_path.open('wb') as buff:
            shutil.copyfileobj(archive.file, buff)
        temp_archive_path.chmod(0o775)
        os.rename(temp_archive_path, archive_path)

    return success, msg

@app.get(plugins_endpoint + '/{name}', tags=['Plugins'])
def get_plugin(
        name: str,
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ) -> Any:
    """
    Download a plugin's archive file.
    """
    plugin = Plugin(name)
    if plugin.archive_path.exists():
        return FileResponse(plugin.archive_path, filename=f'{plugin.name}.tar.gz')
    return False, f"Archive for plugin '{plugin}' could not be found."


@app.get(plugins_endpoint + '/{name}/attributes', tags=['Plugins'])
def get_plugin_attributes(
        name: str,
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ) -> Dict[str, Any]:
    """
    Get a plugin's attributes.
    """
    return get_api_connector().get_plugin_attributes(Plugin(name))

@app.get(plugins_endpoint, tags=['Plugins'])
def get_plugins(
        user_id : Optional[int] = None,
        search_term : Optional[str] = None,
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ) -> List[str]:
    """
    Get a list of plugins.

    Parameters
    ----------
    user_id: Optional[int], default None
        The `user_id` to search by.

    search_term : Optional[str], default None
        Search for plugins with this search term.

    Returns
    -------
    A list of strings.
    """
    return get_api_connector().get_plugins(user_id=user_id, search_term=search_term)

@app.delete(plugins_endpoint + '/{name}', tags=['Plugins'])
def delete_plugin(
        name : str,
        curr_user = (
            fastapi.Depends(manager) if private else None
        ),
    ) -> SuccessTuple:
    """
    Delete a plugin and its archive file from the repository.
    """
    get_tables()
    plugin = Plugin(name)
    plugin_user_id = get_api_connector().get_plugin_user_id(plugin)
    if plugin_user_id is None:
        return False, f"Plugin '{plugin}' is not registered."

    if curr_user is not None:
        curr_user_id = get_api_connector().get_user_id(curr_user)
        if plugin_user_id != curr_user_id:
            return False, f"User '{curr_user.username}' cannot delete plugin '{plugin}'."
    else:
        curr_user_id = -1
    plugin.user_id = curr_user_id

    _remove_success = plugin.remove_archive(debug=debug)
    if not _remove_success[0]:
        return _remove_success

    _delete_success = get_api_connector().delete_plugin(plugin, debug=debug)
    if not _delete_success[0]:
        return _delete_success

    return True, f"Successfully deleted plugin '{plugin}'."



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Define syncing components.
"""

from __future__ import annotations
from meerschaum.api import CHECK_UPDATE
from meerschaum.utils.packages import attempt_import, import_dcc, import_html
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Callbacks for the login page.
"""

from __future__ import annotations
from meerschaum.api import endpoints, no_auth, CHECK_UPDATE
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
import uuid
from meerschaum.utils.typing import Optional
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State
from meerschaum.api.dash import dash_app, debug, pipes, _get_pipes, active_sessions
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.api.routes._login import login
from fastapi_login.exceptions import InvalidCredentialsException
from fastapi.exceptions import HTTPException
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)

@dash_app.callback(
    Output('user-registration-disabled-collapse', 'is_open'),
    Input('show-user-registration-disabled-button', 'n_clicks'),
    State('user-registration-disabled-collapse', 'is_open'),
)
def show_registration_disabled_collapse(n_clicks, is_open):
    """
    Toggle the registration info div.
    """
    if n_clicks:
        return not is_open
    return is_open

@dash_app.callback(
    Output('session-store', 'data'),
    Output('username-input', 'className'),
    Output('location', 'pathname'),
    Input('username-input', 'n_submit'),
    Input('password-input', 'n_submit'),
    Input('login-button', 'n_clicks'),
    State('username-input', 'value'),
    State('password-input', 'value'),
    State('location', 'href'),
)
def login_button_click(
        username_submit,
        password_submit,
        n_clicks,
        username,
        password,
        location_href,
    ):
    """
    When the user submits the login form, check the login.
    On successful login, set the session id.
    """
    form_class = 'form-control'
    ctx = dash.callback_context
    if not username or not password or not ctx.triggered:
        return {}, form_class, dash.no_update
    try:
        token_dict = login({'username': username, 'password': password})
        session_data = {'session-id': str(uuid.uuid4()), 'location.href': location_href}
        active_sessions[session_data['session-id']] = {'username': username}
    except HTTPException:
        form_class += ' is-invalid'
        session_data = None
    return session_data, form_class, (dash.no_update if not session_data else endpoints['dash'])
